from typing import Iterator

import jax.numpy as jnp
import pytest

from cmonge.datasets.single_loader import get_train_valid_test_split


class TestTrainValidTestSplits:
    @pytest.fixture
    def array(self) -> jnp.ndarray:
        return jnp.array([[1], [2], [3], [4], [5]])

    def test_get_train_valid_test_split_normal(self, array):
        split = [0.6, 0.2, 0.2]

        train, valid, test = get_train_valid_test_split(array, split)

        assert train.shape == (3, 1)
        assert valid.shape == (1, 1)
        assert test.shape == (1, 1)

    def test_get_train_valid_test_split_no_train(self, array):
        split = [0, 0.8, 0.2]

        train, valid, test = get_train_valid_test_split(array, split)

        assert train.shape == (0,)
        assert valid.shape == (4, 1)
        assert test.shape == (1, 1)

    def test_get_train_valid_test_split_no_valid(self, array):
        split = [0, 0, 1]

        train, valid, test = get_train_valid_test_split(array, split)

        assert train.shape == (0,)
        assert valid.shape == (0,)
        assert test.shape == (5, 1)

    def test_get_train_valid_test_split_no_test(self, array):
        split = [0, 1, 0]

        train, valid, test = get_train_valid_test_split(array, split)

        assert train.shape == (0,)
        assert valid.shape == (5, 1)
        assert test.shape == (0,)


class TestSyntheticDosageModule:
    def test_train_loaders(self, synthetic_data):
        loaders = synthetic_data.train_dataloaders()
        assert isinstance(loaders[0], Iterator)
        assert isinstance(next(loaders[0]), jnp.ndarray)



import importlib
from pathlib import Path

import pytest

base_dir = Path(__file__).parents[1]
# Skip these directories and files
skip_dirs_files = {"__pycache__", "configs", "tests"}


def discover_modules(path, skip_dirs_files):
    """
    Recursively discovers Python modules in the given path, excluding specified directories and files.
    """
    modules = []
    for item in path.iterdir():
        if item.name in skip_dirs_files or item.name.startswith("_"):
            continue
        if item.is_dir():
            modules += discover_modules(item, skip_dirs_files)
        elif item.suffix == ".py":
            module_path = (
                item.relative_to(base_dir.parent)
                .with_suffix("")
                .as_posix()
                .replace("/", ".")
            )
            modules.append(module_path)
    return modules


# Perform module discovery directly
modules_to_test = discover_modules(base_dir, skip_dirs_files)


def import_module(module_path):
    """
    Attempts to import a module and returns an error message if it fails.
    """
    try:
        importlib.import_module(module_path)
        return None
    except Exception as e:
        return str(e)


@pytest.mark.parametrize("module_path", modules_to_test)
def test_module_import(module_path):
    """
    Test each module import, failing if any module cannot be imported.
    """
    error_message = import_module(module_path)
    assert error_message is None, f"Failed to import {module_path}: {error_message}"



from pathlib import Path
from typing import Literal

import jax
import jax.numpy as jnp
import optax
from dotmap import DotMap
from flax.core.scope import FrozenVariableDict
from loguru import logger
from ott.neural.networks.icnn import ICNN

from cmonge.datasets.conditional_loader import ConditionalDataModule
from cmonge.evaluate import init_logger_dict, log_mean_metrics, log_metrics
from cmonge.models.nn import PICNN
from cmonge.trainers.ot_trainer import AbstractTrainer
from cmonge.utils import create_or_update_logfile


class ConditionalTrainer(AbstractTrainer):
    """Wrapper class for conditional neural dual training."""

    def __init__(
        self, jobid: int, logger_path: Path, config: DotMap, datamodule
    ) -> None:
        super().__init__(jobid, logger_path)
        self.init_model(datamodule=datamodule, **config)
        self.create_functions()

    def init_model(
        self,
        num_train_iters: int,
        num_inner_iters: int,
        lr: float,
        seed: int,
        num_genes: int,
        dim_hidden: list[int],
        cond_dim: int,
        embedding: str,
        method: str,
        pos_weights: bool,
        datamodule,
    ):
        self.num_inner_iters = num_inner_iters
        self.num_train_iters = num_train_iters
        self.lr = lr
        self.seed = seed
        self.num_genes = num_genes
        self.dim_hidden = dim_hidden
        self.num_genes = self.input_dim = num_genes
        self.cond_dim = cond_dim
        self.embedding = embedding
        self.method = method
        self.pos_weights = pos_weights
        self.beta = 1
        loaders = datamodule.train_dataloaders()
        self.generate_embeddings(datamodule)
        w_0 = self.init_modulator_weights(loaders)
        factors, means = self.compute_gaussian_map_params(loaders)

        self.key = jax.random.PRNGKey(self.seed)
        self.key, key_f, key_g = jax.random.split(self.key, 3)

        self.neural_f = PICNN(
            dim_hidden=self.dim_hidden,
            dim_data=self.num_genes,
            cond_dim=self.cond_dim,
            conditions=w_0,
            factors=factors,
            means=means,
        )
        self.neural_g = PICNN(
            dim_hidden=self.dim_hidden,
            dim_data=self.num_genes,
            cond_dim=self.cond_dim,
            conditions=w_0,
            factors=factors,
            means=means,
        )

        lr_schedule = optax.cosine_decay_schedule(
            init_value=self.lr, decay_steps=self.num_inner_iters, alpha=1e-2
        )
        optimizer_f = optax.adamw(learning_rate=lr_schedule)
        optimizer_g = optax.adamw(learning_rate=lr_schedule)

        self.state_f = self.neural_f.create_train_state(
            key_f,
            optimizer_f,
            self.num_genes,
        )
        self.state_g = self.neural_g.create_train_state(
            key_g,
            optimizer_g,
            self.num_genes,
        )

    def compute_gaussian_map_params(self, train_loaders):
        factors = []
        means = []
        for cond, loaders in train_loaders.items():
            samples = next(loaders[0]), next(loaders[1])
            factor, mean = ICNN._compute_gaussian_map_params(samples)

            factors.append(factor)
            means.append(mean)
        return jnp.vstack(factors), jnp.vstack(means)

    def init_modulator_weights(self, train_loaders):
        conditions = []
        for cond, loaders in train_loaders.items():
            embedding = self.embeddings[cond]
            conditions.append(jnp.asarray(embedding))
        return jnp.stack(conditions).T

    def _penalize_weights_icnn(self, params: FrozenVariableDict) -> float:
        """Penalize weights of ICNN."""
        penalty = 0
        for key in params:
            if key.startswith("wz_"):
                penalty += jnp.linalg.norm(jax.nn.relu(-params[key]["kernel"]))
        return penalty

    def _clip_weights_icnn(self, params: FrozenVariableDict) -> FrozenVariableDict:
        """Clip weights of ICNN."""
        for key in params:
            if key.startswith("wz_"):
                params[key]["kernel"] = jnp.clip(params[key]["kernel"], a_min=0)

        return params

    def get_step_fn(self, train: bool, to_optimize: Literal["f", "g"]):
        """Create a parallel training and evaluation function."""

        def loss_fn(params_f, params_g, f_value, g_value, g_gradient, batch):
            """Loss function for both potentials."""
            source, target, condition = (
                batch["source"],
                batch["target"],
                batch["condition"],
            )

            init_source_hat = g_gradient(params_g)(target, condition)
            batch_dot = jax.vmap(jnp.dot)

            f_source = f_value(params_f)(source, condition)
            f_star_target = batch_dot(init_source_hat, target) - f_value(params_f)(
                init_source_hat, condition
            )
            dual_source = f_source.mean()
            dual_target = f_star_target.mean()
            loss_f = dual_source + dual_target

            f_value_parameters_detached = f_value(jax.lax.stop_gradient(params_f))
            loss_g = (
                f_value_parameters_detached(init_source_hat, condition)
                - batch_dot(init_source_hat, target)
            ).mean()

            # compute Wasserstein-2 distance
            c = jnp.mean(jnp.sum(source**2, axis=-1)) + jnp.mean(
                jnp.sum(target**2, axis=-1)
            )
            w2_dist = c - 2.0 * (f_source.mean() + f_star_target.mean())

            if not self.pos_weights:
                penalty = self.beta * self._penalize_weights_icnn(
                    params_f
                ) + self.beta * self._penalize_weights_icnn(params_g)
                loss_f += penalty
                loss_g += penalty
            else:
                penalty = 0

            if to_optimize == "f":
                return loss_f, (w2_dist, loss_f, loss_g, penalty)
            if to_optimize == "g":
                return loss_g, (w2_dist, loss_f, loss_g, penalty)

        @jax.jit
        def step_fn(state_f, state_g, batch):
            """Step function of either training or validation."""
            grad_fn = jax.value_and_grad(loss_fn, argnums=[0, 1], has_aux=True)
            if train:
                # compute loss and gradients
                (_, (w2_dist, loss_f, loss_g, penalty)), (grads_f, grads_g) = grad_fn(
                    state_f.params,
                    state_g.params,
                    state_f.potential_value_fn,
                    state_g.potential_value_fn,
                    state_f.potential_gradient_fn,
                    batch,
                )

                if to_optimize == "f":
                    return (
                        state_f.apply_gradients(grads=grads_f),
                        w2_dist,
                        loss_f,
                        loss_g,
                        penalty,
                    )
                if to_optimize == "g":
                    return (
                        state_g.apply_gradients(grads=grads_g),
                        w2_dist,
                        loss_f,
                        loss_g,
                        penalty,
                    )
                raise ValueError("Optimization target has been misspecified.")

            # compute loss and gradients
            (_, (w2_dist, loss_f, loss_g, penalty)), _ = grad_fn(
                state_f.params,
                state_g.params,
                state_f.potential_value_fn,
                state_g.potential_value_fn,
                state_f.potential_gradient_fn,
                batch,
            )

            return w2_dist, loss_f, loss_g, penalty

        return step_fn

    def create_functions(self):
        self.train_step_f = self.get_step_fn(train=True, to_optimize="f")
        self.train_step_g = self.get_step_fn(train=True, to_optimize="g")
        self.valid_step_f = self.get_step_fn(train=False, to_optimize="f")
        self.valid_step_g = self.get_step_fn(train=False, to_optimize="g")

    def train(self, datamodule: ConditionalDataModule, valid: bool = False):
        logger.info("Training started.")
        condition_to_loaders = datamodule.train_dataloaders()

        self.generate_embeddings(datamodule)

        # self.pretrain_identity(datamodule)

        batch_g, batch_f, valid_batch = {}, {}, {}
        valid_logs = {"loss_f": [], "loss_g": [], "w_dist": []}

        #     return pretrain_logs
        train_logs = {"loss_f": [], "loss_g": [], "w_dist": []}
        for step in range(self.num_train_iters):
            condition = datamodule.sample_condition("train")
            trainloader_source, trainloader_target = condition_to_loaders[condition]
            condition = self.embeddings[condition]
            condition_batch = jnp.asarray(
                [condition for _ in range(datamodule.batch_size)]
            )

            for _ in range(self.num_inner_iters):
                batch_g["source"] = jnp.asarray(next(trainloader_source))
                batch_g["target"] = jnp.asarray(next(trainloader_target))
                batch_g["condition"] = condition_batch

                self.state_g, w_dist, loss_f, loss_g, penalty = self.train_step_g(
                    self.state_f, self.state_g, batch_g
                )

            batch_f["source"] = jnp.asarray(next(trainloader_source))
            batch_f["target"] = jnp.asarray(next(trainloader_target))
            batch_f["condition"] = condition_batch

            self.state_f, w_dist, loss_f, loss_g, penalty = self.train_step_f(
                self.state_f, self.state_g, batch_f
            )

            if valid and step % 100 == 0 and step > 0:
                condition = datamodule.sample_condition("valid")
                valid_condition_to_loaders = datamodule.valid_dataloaders()
                validloader_source, validloader_target = valid_condition_to_loaders[
                    condition
                ]
                condition = self.embeddings[condition]
                condition_batch = jnp.asarray(
                    [condition for _ in range(datamodule.batch_size)]
                )

                valid_batch["source"] = jnp.asarray(next(validloader_source))
                valid_batch["target"] = jnp.asarray(next(validloader_target))
                valid_batch["condition"] = condition_batch

                valid_w_dist, valid_loss_f, valid_loss_g, penalty = self.valid_step_g(
                    self.state_f, self.state_g, valid_batch
                )

                self.update_logging(train_logs, loss_f, loss_g, w_dist, "train", step)
                self.update_logging(
                    valid_logs, valid_loss_f, valid_loss_g, valid_w_dist, "valid", step
                )

        self.metrics["ottlogs"] = {"train_logs": train_logs, "valid_logs": valid_logs}
        logger.info("Training finished.")

    def update_logging(self, log_dict, loss_f, loss_g, w_dist, kind, step):
        log_dict["loss_f"].append(loss_f)
        log_dict["loss_g"].append(loss_g)
        log_dict["w_dist"].append(w_dist)
        loss_f_format = "{:.2f}".format(loss_f)
        loss_g_format = "{:.2f}".format(loss_g)
        w_dist_fomrat = "{:.2f}".format(w_dist)
        logger.info(
            f"{step} n_iters - {kind} - loss_f: {loss_f_format}, loss_g: {loss_g_format}, w_dist: {w_dist_fomrat}"
        )
        return log_dict

    def generate_embeddings(self, datamodule: ConditionalDataModule):
        assert self.embedding in ["dosage", "dense"]
        if self.embedding == "dosage":
            embeddings = {
                cond: [int(cond.split("-")[1]) / 10000]
                for cond in datamodule.conditions
            }
            self.embeddings = embeddings
            datamodule.embeddings = embeddings

    def transport(self, x: jnp.ndarray, condition: jnp.ndarray) -> jnp.ndarray:
        """Conditionally transport according to Brenier formula."""

        def f(x, c) -> float:
            return self.state_g.apply_fn({"params": self.state_g.params}, x, c)

        return jax.vmap(jax.grad(f, argnums=0))(x, condition)

    def evaluate(
        self,
        datamodule: ConditionalDataModule,
        identity: bool = False,
        n_samples: int = 9,
    ) -> None:
        """Evaluate a trained model on a validation set and save the metrics to a json file."""

        def evaluate_condition(loader_source, loader_target, cond_embeddings, metrics):
            for enum, (source, target) in enumerate(zip(loader_source, loader_target)):
                if not identity:
                    transport = self.transport(source, cond_embeddings)

                else:
                    transport = source

                transport = datamodule.decoder(transport)
                target = datamodule.decoder(target)

                if datamodule.marker_idx:
                    target = target[:, datamodule.marker_idx]
                    transport = transport[:, datamodule.marker_idx]

                log_metrics(metrics, target, transport)
                if enum > n_samples:
                    break

        cond_to_loaders = datamodule.train_dataloaders()
        for cond, loader in cond_to_loaders.items():
            logger.info(f"Evaluation started on {cond} (in-sample).")
            cond_embedding = self.embeddings[cond]
            cond_embedding = jnp.asarray(
                [[cond_embedding] for _ in range(datamodule.batch_size)]
            )
            loader_source, loader_target = loader

            self.metrics[f"in-sample-{cond}"] = {}
            init_logger_dict(
                self.metrics[f"in-sample-{cond}"], datamodule.drug_condition
            )
            evaluate_condition(
                loader_source,
                loader_target,
                cond_embedding,
                self.metrics[f"in-sample-{cond}"],
            )
            log_mean_metrics(self.metrics[f"in-sample-{cond}"])

        cond_to_loaders = datamodule.valid_dataloaders()
        for cond, loader in cond_to_loaders.items():
            logger.info(f"Evaluation started on {cond} (out-sample).")
            cond_embedding = self.embeddings[cond]
            cond_embedding = jnp.asarray(
                [[cond_embedding] for _ in range(datamodule.batch_size)]
            )
            loader_source, loader_target = loader

            self.metrics[f"out-sample-{cond}"] = {}
            init_logger_dict(
                self.metrics[f"out-sample-{cond}"], datamodule.drug_condition
            )
            evaluate_condition(
                loader_source,
                loader_target,
                cond_embedding,
                self.metrics[f"out-sample-{cond}"],
            )
            log_mean_metrics(self.metrics[f"out-sample-{cond}"])

        create_or_update_logfile(self.logger_path, self.metrics)

        create_or_update_logfile(self.logger_path, self.metrics)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Callable, Sequence

import jax
import jax.numpy as jnp
import numpy as np
import optax
from dotmap import DotMap
from flax import linen as nn
from flax.core.frozen_dict import FrozenDict
from flax.training import checkpoints
from flax.training.train_state import TrainState
from loguru import logger

from cmonge.utils import activation_factory, optim_factory

if TYPE_CHECKING:
    from cmonge.datasets.single_loader import AbstractDataModule


class Encoder(nn.Module):
    """Flax Encoder module."""

    hidden_dims: Sequence[int]
    latent_dim: int
    act_fn: Callable[[jnp.ndarray], jnp.ndarray] = nn.gelu

    @nn.compact
    def __call__(self, x):
        for n_hidden in self.hidden_dims:
            w = nn.Dense(n_hidden, use_bias=True)
            x = self.act_fn(w(x))
        x = nn.Dense(features=self.latent_dim)(x)
        return x


class Decoder(nn.Module):
    """Flax Decoder module."""

    hidden_dims: Sequence[int]
    data_dim: int
    act_fn: Callable[[jnp.ndarray], jnp.ndarray] = nn.gelu

    @nn.compact
    def __call__(self, x):
        for n_hidden in self.hidden_dims:
            w = nn.Dense(n_hidden, use_bias=True)
            x = self.act_fn(w(x))
        x = nn.Dense(features=self.data_dim)(x)
        return x


class AutoEncoder(nn.Module):
    """Flax AutoEncoder module."""

    hidden_dims: Sequence[int]
    latent_dim: int
    data_dim: int
    seed: int
    act_fn: Callable[[jnp.ndarray], jnp.ndarray] = nn.gelu

    def setup(self):
        self.encoder = Encoder(self.hidden_dims, self.latent_dim, self.act_fn)
        self.decoder = Decoder(self.hidden_dims, self.data_dim, self.act_fn)

    def __call__(self, x):
        z = self.encoder(x)
        x_hat = self.decoder(z)
        return x_hat


def mse_reconstruction_loss(model: AutoEncoder, params: FrozenDict, batch: jnp.ndarray):
    """l2 reconstruction loss for the autoencoder."""
    pred = model.apply({"params": params}, batch)
    return 2 * optax.l2_loss(pred, batch).mean()


class AETrainerModule:
    """Class for training, evaluating and saving an AutoEncoder model."""

    def __init__(self, config: DotMap):
        self.config = config
        self.model_dir = Path(self.config.training.model_dir)
        self.create_functions()
        self.init_model()

    def init_model(self):
        """Initialize optimizer and model parameters."""
        self.config.model.act_fn = activation_factory[self.config.model.act_fn]
        self.model = AutoEncoder(**self.config.model)

        rng = jax.random.PRNGKey(self.config.model.seed)
        rng, init_rng = jax.random.split(rng)
        empty_array = jnp.empty((1, self.config.model.data_dim))
        self.latent_shift = empty_array
        params = self.model.init(init_rng, empty_array)["params"]

        lr_scheduler = optax.warmup_cosine_decay_schedule(
            init_value=0.0,
            peak_value=self.config.optim.lr,
            warmup_steps=100,
            decay_steps=self.config.training.n_epochs * 100,
            end_value=1e-5,
        )
        opt_fn = optim_factory[self.config.optim.optimizer]
        optimizer = opt_fn(learning_rate=lr_scheduler, **self.config.optim.kwargs)

        self.state = TrainState.create(
            apply_fn=self.model.apply, params=params, tx=optimizer
        )

    def create_functions(self):
        """Define jitted train and eval step on a batch of input."""

        def train_step(state: TrainState, batch: jnp.ndarray):

            def loss_fn(params):
                return mse_reconstruction_loss(self.model, params, batch)

            loss, grads = jax.value_and_grad(loss_fn)(
                state.params
            )  # Get loss and gradients for loss
            state = state.apply_gradients(grads=grads)  # Optimizer update step
            return state, loss

        def eval_step(state: TrainState, batch: jnp.ndarray):
            return mse_reconstruction_loss(self.model, state.params, batch)

        self.train_step = jax.jit(train_step)
        self.eval_step = jax.jit(eval_step)

    def generate_batch(
        self,
        datamodule: ConditionalDataModule,
        split_type: str,
    ) -> Dict[str, jnp.ndarray]:
        """Generate a batch of condition and samples."""
        condition = datamodule.sample_condition(split_type)
        loaders = datamodule.loaders[condition]
        loader = loaders.get_loaders_by_type(split_type)
        return next(loader)

    def train(self, datamodule: AbstractDataModule):
        """Train model for n_epochs, save best model after each epoch."""
        logger.info("Training started.")
        best_eval = 1e6
        for epoch in range(self.config.training.n_epochs):
            # Training step
            batch = self.generate_batch(datamodule, "train")
            self.state, loss = self.train_step(state=self.state, batch=batch)
            logger.info(f"train/loss - epoch {epoch}: {loss}")

            # Validation step
            if self.config.training.valid:
                batch = self.generate_batch(datamodule, "train")
                loss = self.eval_step(self.state, batch)
                # batch_sizes = batch.shape[0]
                # eval_loss = (loss * batch_size).sum() / batch_size.sum()
                eval_loss = loss
                logger.info(f"valid/loss - epoch {epoch}: {loss}")

                # Saving checkpoint if eval metric improved
                if eval_loss < best_eval and self.config.training.cpkt:
                    best_eval = eval_loss
                    self.compute_latent_shift(datamodule)
                    self.save_model(
                        dataset_name=datamodule.name,
                        drug_condition=datamodule.drug_condition,
                        step=epoch,
                    )

        # save the final model if no checkpointing
        if not self.config.training.cpkt:
            self.save_model(
                dataset_name=datamodule.name,
                drug_condition=datamodule.drug_condition,
                step=epoch,
            )
        logger.info("Training finished.")

    def save_model(self, dataset_name: str, drug_condition: str, step: int = 0):
        """Save current model at certain training iteration."""
        model_dir = self.model_dir / dataset_name
        model_dir.mkdir(parents=True, exist_ok=True)
        cpkt = {"params": self.state.params, "latent_shift": self.latent_shift}
        prefix = f"autoencoder_{self.model.latent_dim}_{dataset_name}_{drug_condition}_"
        logger.info(f"Saving model checkpoint to {prefix}")

        checkpoints.save_checkpoint(
            ckpt_dir=model_dir,
            target=cpkt,
            prefix=prefix,
            step=step,
            overwrite=True,
        )
        logger.info("Model checkpoint saved.")

    def load_model(self, dataset_name: str, drug_condition: str):
        """Load model from checkpoint."""
        model_dir = self.model_dir / dataset_name
        target = {"params": self.state.params, "latent_shift": self.latent_shift}
        prefix = f"autoencoder_{self.model.latent_dim}_{dataset_name}_{drug_condition}_"
        logger.info(f"Loading AE model checkpoint from {model_dir}, {prefix}")
        cpkt = checkpoints.restore_checkpoint(
            ckpt_dir=model_dir,
            prefix=prefix,
            target=target,
        )
        if cpkt["params"] is self.state.params:
            logger.info("Failed to load AE model checkpoint.")
        else:
            logger.info("AE Model checkpoint loaded.")
            logger.info(f"AE Model: {prefix}")

        self.state = TrainState.create(
            apply_fn=self.model.apply, params=cpkt["params"], tx=self.state.tx
        )
        self.latent_shift = cpkt["latent_shift"]

    def compute_latent_shift(self, datamodule: AbstractDataModule):
        logger.info(f"Computing latent shift for drug {datamodule.drug_condition}.")
        source = datamodule.adata[datamodule.control_train_cells, :].X
        target = datamodule.adata[datamodule.target_train_cells, :].X
        self.model_bd = self.model.bind({"params": self.state.params})
        mean_encoded_source = self.model_bd.encoder(source).mean(0)
        mean_encoded_target = self.model_bd.encoder(target).mean(0)
        self.latent_shift = mean_encoded_target - mean_encoded_source
        logger.info("Latent shift computed.")

    def transport(self, source: jnp.ndarray) -> jnp.ndarray:
        self.model_bd = self.model.bind({"params": self.state.params})
        z = self.model_bd.encoder(source)
        z_hat = z + self.latent_shift
        transport = self.model_bd.decoder(z_hat)
        return transport



import json
import os
import random
from pathlib import Path
from typing import Any, Dict

import jax.numpy as jnp
import optax
import pandas as pd
import yaml
from dotmap import DotMap
from flax import linen as nn


def monge_get_source_target_transport(
    trainer: "MongeGapTrainer",
    datamodule: "AbstractDataModule",
    n_samples=1,
    target=True,
    source=True,
    transport=True,
    batch_size: int = None,
    num_contexts: int = 2,
):
    if batch_size is None:
        batch_size = datamodule.batch_size

    if datamodule.split[2] > 0:
        print("Evaluating on test set")
        split_type = "test"
    elif datamodule.split[1] > 0:
        print("Evaluating on validation set")
        split_type = "valid"
    else:
        print("Evaluating on training set")
        split_type = "train"
    all_expr = []
    all_meta = []
    for i in range(n_samples):
        if split_type == "valid":
            sel_target_cells = random.sample(
                datamodule.target_valid_cells.tolist(),
                batch_size,
            )
            sel_control_cells = random.sample(
                datamodule.control_valid_cells.tolist(),
                batch_size,
            )

        elif split_type == "test":
            sel_target_cells = random.sample(
                datamodule.target_test_cells.tolist(), batch_size
            )
            sel_control_cells = random.sample(
                datamodule.control_test_cells.tolist(),
                batch_size,
            )
        elif split_type == "train":
            sel_target_cells = random.sample(
                datamodule.target_train_cells.tolist(),
                batch_size,
            )
            sel_control_cells = random.sample(
                datamodule.control_train_cells.tolist(),
                batch_size,
            )

        cond_expr = datamodule.adata[sel_target_cells].X
        cond_meta = datamodule.adata.obs.loc[sel_target_cells, :]

        source_expr = datamodule.adata[sel_control_cells].X
        source_meta = datamodule.adata.obs.loc[sel_control_cells, :]

        if target:
            cond_meta["sample_n"] = i
            cond_meta["dtype"] = "target"
            all_expr.append(pd.DataFrame(cond_expr, columns=datamodule.adata.var_names))
            all_meta.append(cond_meta)

        if source:
            source_meta["dtype"] = "source"
            source_meta["sample_n"] = i

            all_meta.append(source_meta)
            all_expr.append(
                pd.DataFrame(source_expr, columns=datamodule.adata.var_names)
            )

        if transport:
            if num_contexts == 0:
                trans = trainer.transport(source_expr)
            else:
                trans = trainer.transport(source_expr, num_contexts=num_contexts)
            trans = datamodule.decoder(trans)
            trans_meta = cond_meta.copy()
            trans_meta["dtype"] = "transport"
            trans_meta["sample_n"] = i

            all_expr.append(pd.DataFrame(trans, columns=datamodule.adata.var_names))
            all_meta.append(trans_meta)

    all_expr = pd.concat(all_expr).reset_index(drop=True)
    all_meta = pd.concat(all_meta).reset_index(drop=True)

    return all_expr, all_meta


def get_source_target_transport(
    trainer: "ConditionalMongeTrainer",
    datamodule: "ConditionalDataModule",
    conditions,
    target=True,
    source=True,
    transport=True,
):
    if datamodule.data_config.split[2] > 0:
        print("Evaluating on test set")
        split_type = "test"

    elif datamodule.data_config.split[1] > 0:
        print("Evaluating on validation set")
        split_type = "valid"

    else:
        print("Evaluating on training set")
        split_type = "train"

    all_expr = []
    for condition in conditions:
        print(condition)
        cond_embeddings, num_contexts = trainer.embedding_module(condition)

        if split_type == "valid":
            loader_source, loader_target = datamodule.valid_dataloaders()[condition]
        elif split_type == "test":
            loader_source, loader_target = datamodule.test_dataloaders()[condition]
        elif split_type == "train":
            loader_source, loader_target = datamodule.train_dataloaders()[condition]

        target_expr = next(iter(loader_target))
        source_expr = next(iter(loader_source))

        if target:
            res_target = datamodule.decoder(target_expr)
            res_target = pd.DataFrame(res_target)
            res_target["dtype"] = "target"
            res_target["condition"] = condition
            all_expr.append(res_target)

        if source:
            res_source = datamodule.decoder(source_expr)
            res_source = pd.DataFrame(res_source)
            res_source["dtype"] = "source"
            res_source["condition"] = condition
            all_expr.append(res_source)

        if transport:
            trans = trainer.transport(source_expr, cond_embeddings, num_contexts)
            trans = datamodule.decoder(trans)
            trans = pd.DataFrame(trans)
            trans["dtype"] = "trans"
            trans["condition"] = condition
            all_expr.append(trans)

    return pd.concat(all_expr)


def jax_serializer(obj):
    if isinstance(obj, jnp.ndarray):
        return obj.tolist()
    raise TypeError("Type not serializable")


def create_or_update_logfile(file_path: Path, item: Dict[str, Any] = {}):
    """Updates the logfile containing the model experimetns."""
    if not os.path.exists(file_path):
        with open(file_path, "w") as file:
            json.dump({"experiments": []}, file)

    with open(file_path, "r") as file:
        data = json.load(file)

    data["experiments"].append(item)
    with open(file_path, "w") as file:
        json.dump(data, file, indent=4, default=jax_serializer)


def load_config(path: Path) -> DotMap[str, Any]:
    """Load model and data configs from yaml file."""

    with open(path, "r") as file:
        yaml_data = yaml.safe_load(file)
    yaml_data = DotMap(yaml_data)
    return yaml_data


optim_factory = {"adamw": optax.adamw, "adam": optax.adam, "sgd": optax.sgd}
activation_factory = {"gelu": nn.gelu, "relu": nn.relu, "leakyrelu": nn.leaky_relu}



from typing import Any, Callable, Iterable, Sequence, Tuple, Union

import flax.linen as nn
import jax
import jax.numpy as jnp
import optax
from flax.core import frozen_dict
from jax.nn import initializers
from ott.neural.networks.icnn import ICNN
from ott.neural.networks.layers.posdef import PosDefPotentials
from ott.neural.networks.potentials import (
    BasePotential,
    PotentialGradientFn_t,
    PotentialTrainState,
    PotentialValueFn_t,
)


class PICNN(ICNN):
    """Partial Input convex neural network (PICNN) architecture."""

    dim_data: int = None
    dim_hidden: Sequence[int] = None
    cond_dim: int = None
    init_std: float = 1e-2
    init_fn: Callable = jax.nn.initializers.normal
    act_fn: Callable[[jnp.ndarray], jnp.ndarray] = nn.leaky_relu
    pos_weights: bool = True
    conditions: jnp.ndarray = None
    factors: jnp.ndarray = None
    means: jnp.ndarray = None

    def setup(self):
        self.num_hidden = len(self.dim_hidden)
        self.input_dim = self.dim_data
        self.num_conditions = len(self.means)
        units = [self.num_conditions] + list(self.dim_hidden)
        self.n_layers = len(units)

        # self layers for hidden state u, when contributing all ~0
        w = []
        for odim in [units[0]] + units[:-1]:
            _w = nn.Dense(
                odim,
                use_bias=True,
                kernel_init=self.init_fn(self.init_std),
                bias_init=self.init_fn(self.init_std),
            )
            w.append(_w)
        self.w = w

        # first layer for hidden state performs a comparison with database
        self.w_0 = self.conditions

        # auto layers for z, should be mean operators with no bias
        # keep track of previous size to normalize accordingly
        wz = []
        # keep track of previous size to normalize accordingly
        normalization = 1

        for odim in units[1:] + [1]:
            wz.append(
                nn.Dense(
                    odim,
                    kernel_init=initializers.constant(1.0 / normalization),
                    use_bias=False,
                )
            )
            normalization = odim
        self.wz = wz

        # for family of convex functions stored in z, if using init then first
        # vector z_0 has as many values as # of convex potentials.
        w_z0 = PosDefPotentials(
            dim_data=self.dim_data,
            num_potentials=self.num_conditions,
            use_bias=True,
            kernel_init=lambda *_: self.factors,
            bias_init=lambda *_: self.means,
        )
        self.w_z0 = w_z0

        # cross layers for convex functions z / hidden state u
        # initialized to be identity first with 0 bias
        # and then ~0 + 1 bias to ensure identity
        wzu = []
        _wzu = nn.Dense(
            units[0],
            use_bias=True,
            kernel_init=self.init_fn(self.init_std),
            bias_init=initializers.constant(1.0),
        )
        wzu.append(_wzu)

        for odim in units[1:]:
            _wzu = nn.Dense(
                odim,
                use_bias=True,
                kernel_init=self.init_fn(self.init_std),
                bias_init=initializers.constant(1.0),
            )
            wzu.append(_wzu)
        self.wzu = wzu

        # self layers for x, ~0
        wx = []
        for odim in units + [1]:
            _wx = nn.Dense(
                odim,
                use_bias=True,
                kernel_init=self.init_fn(self.init_std),
                bias_init=initializers.constant(0.0),
            )
            wx.append(_wx)
        self.wx = wx

        # cross layers for x / hidden state u, all ~0
        wxu = []
        for idim in [units[0]] + units:
            _wxu = nn.Dense(
                self.dim_data,
                use_bias=True,
                kernel_init=self.init_fn(self.init_std),
                bias_init=initializers.constant(0.0),
            )
            wxu.append(_wxu)
        self.wxu = wxu

        # self layers for hidden state u, to update z, all ~0
        wu = []
        for odim in units + [1]:
            _wu = nn.Dense(
                odim, use_bias=False, kernel_init=self.init_fn(self.init_std)
            )
            wu.append(_wu)
        self.wu = wu

    @nn.compact
    def __call__(self, x: jnp.ndarray, c: jnp.ndarray) -> jnp.ndarray:
        u = jax.nn.softmax(10 * c @ self.w_0)
        z_0 = self.w_z0(x)
        z = self.act_fn(z_0 * u)
        # apply k layers - 1
        for i in range(1, self.n_layers):
            u = self.act_fn(self.w[i](u))
            t_u = jax.nn.softplus(self.wzu[i - 1](u))
            z = self.act_fn(
                self.wz[i - 1](jnp.multiply(z, t_u))
                + self.wx[i](jnp.multiply(x, self.wxu[i](u)))
                + self.wu[i](u)
            )

        z = (
            self.wz[-1](jnp.multiply(z, jax.nn.softplus(self.wzu[-1](u))))
            + self.wx[-1](jnp.multiply(x, self.wxu[-1](u)))
            + self.wu[-1](u)
        )
        return z.squeeze()

    def create_train_state(
        self,
        rng: jnp.ndarray,
        optimizer: optax.OptState,
        input_shape: Union[int, Tuple[int, ...]],
        **kwargs: Any,
    ) -> PotentialTrainState:
        """Create initial `TrainState`."""
        condition = jnp.ones((1, self.cond_dim))
        params = self.init(rng, x=jnp.ones((1, input_shape)), c=condition)["params"]
        return PotentialTrainState.create(
            apply_fn=self.apply,
            params=params,
            tx=optimizer,
            potential_value_fn=self.potential_value_fn,
            potential_gradient_fn=self.potential_gradient_fn,
            **kwargs,
        )

    def potential_value_fn(
        self,
        params: frozen_dict.FrozenDict[str, jnp.ndarray],
    ) -> PotentialValueFn_t:
        """A function that can be evaluated to obtain a potential value, or a linear
        interpolation of a potential.
        """
        return lambda x, c: self.apply({"params": params}, x=x, c=c)  # type: ignore[misc]

    def potential_gradient_fn(
        self,
        params: frozen_dict.FrozenDict[str, jnp.ndarray],
    ) -> PotentialGradientFn_t:
        """Return a function returning a vector or the gradient of the potential.
        Args:
        params: parameters of the module
        Returns
        -------
        A function that can be evaluated to obtain the potential's gradient
        """
        return jax.vmap(jax.grad(self.potential_value_fn(params), argnums=0))


class ConditionalMLP(BasePotential):
    dim_hidden: Sequence[int] = None
    dim_data: int = None
    dim_cond: int = None
    act_fn: Callable[[jnp.ndarray], jnp.ndarray] = nn.gelu
    is_potential: bool = False

    @nn.compact
    def __call__(self, x: jnp.ndarray, c: jnp.ndarray) -> jnp.ndarray:  # noqa: D102
        n_input = x.shape[-1]
        z = jnp.concatenate((x, c), axis=1)

        for n_hidden in self.dim_hidden:
            wx = nn.Dense(n_hidden, use_bias=True)
            z = self.act_fn(wx(z))
        wx = nn.Dense(n_input, use_bias=True)
        z = x + wx(z)
        return z

    def create_train_state(
        self,
        rng: jnp.ndarray,
        optimizer: optax.OptState,
        **kwargs: Any,
    ) -> PotentialTrainState:
        """Create initial `TrainState`."""
        c = jnp.ones((1, self.dim_cond))
        x = jnp.ones((1, self.dim_data))
        params = self.init(rng, x=x, c=c)["params"]
        return PotentialTrainState.create(
            apply_fn=self.apply,
            params=params,
            tx=optimizer,
            potential_value_fn=self.potential_value_fn,
            potential_gradient_fn=self.potential_gradient_fn,
            **kwargs,
        )


class DummyMLP(BasePotential):
    """A generic, typically not-convex (w.r.t input) MLP.

    Args:
      dim_hidden: sequence specifying size of hidden dimensions. The output
        dimension of the last layer is automatically set to 1 if
        :attr:`is_potential` is ``True``, or the dimension of the input otherwise
      is_potential: Model the potential if ``True``, otherwise
        model the gradient of the potential
      act_fn: Activation function
    """

    dim_hidden: Sequence[int] = None
    dim_data: int = None
    dim_cond: int = None
    act_fn: Callable[[jnp.ndarray], jnp.ndarray] = nn.gelu
    is_potential: bool = False

    @nn.compact
    def __call__(self, x: jnp.ndarray, c: jnp.ndarray) -> jnp.ndarray:  # noqa: D102
        squeeze = x.ndim == 1
        if squeeze:
            x = jnp.expand_dims(x, 0)
        assert x.ndim == 2, x.ndim
        n_input = x.shape[-1]

        z = x
        for n_hidden in self.dim_hidden:
            wx = nn.Dense(n_hidden, use_bias=True)
            z = self.act_fn(wx(z))

        if self.is_potential:
            wx = nn.Dense(1, use_bias=True)
            z = wx(z).squeeze(-1)

            quad_term = 0.5 * jax.vmap(jnp.dot)(x, x)
            z += quad_term
        else:
            wx = nn.Dense(n_input, use_bias=True)
            z = x + wx(z)

        return z.squeeze(0) if squeeze else z

    def create_train_state(
        self,
        rng: jnp.ndarray,
        optimizer: optax.OptState,
        **kwargs: Any,
    ) -> PotentialTrainState:
        """Create initial `TrainState`."""
        c = jnp.ones((1, self.dim_cond))
        x = jnp.ones((1, self.dim_data))
        params = self.init(rng, x=x, c=c)["params"]
        return PotentialTrainState.create(
            apply_fn=self.apply,
            params=params,
            tx=optimizer,
            potential_value_fn=self.potential_value_fn,
            potential_gradient_fn=self.potential_gradient_fn,
            **kwargs,
        )


class ConditionalPerturbationNetwork(BasePotential):
    dim_hidden: Sequence[int] = None
    dim_data: int = None
    dim_cond: int = None  # Full dimension of all context variables concatenated
    # Same length as context_entity_bonds if embed_cond_equal is False
    # (if True, first item is size of deep set layer, rest is ignored)
    dim_cond_map: Iterable[int] = (50,)
    act_fn: Callable[[jnp.ndarray], jnp.ndarray] = nn.gelu
    is_potential: bool = False
    layer_norm: bool = False
    embed_cond_equal: bool = (
        False  # Whether all context variables should be treated as set or not
    )
    context_entity_bonds: Iterable[Tuple[int, int]] = (
        (0, 10),
        (0, 11),
    )  # Start/stop index per modality

    @nn.compact
    def __call__(
        self, x: jnp.ndarray, c: jnp.ndarray, num_contexts: int = 2
    ) -> jnp.ndarray:  # noqa: D102
        """
        Args:
            x (jnp.ndarray): The input data of shape bs x dim_data
            c (jnp.ndarray): The context of shape bs x dim_cond with
                possibly different modalities
                concatenated, as can be specified via context_entity_bonds.

        Returns:
            jnp.ndarray: _description_
        """
        n_input = x.shape[-1]

        # Chunk the inputs
        contexts = [
            c[:, e[0] : e[1]]
            for i, e in enumerate(self.context_entity_bonds)
            if i < num_contexts
        ]

        # Backwards compatability for dim_cond_map
        if (
            not isinstance(self.dim_cond_map, Iterable)
            and self.dim_cond_map is not None
        ):
            if not self.embed_cond_equal:
                dim_cond_map = [self.dim_cond_map, 1]  # For sciplex backwards compat
            else:
                dim_cond_map = [self.dim_cond_map]
        else:
            dim_cond_map = self.dim_cond_map

        if not self.embed_cond_equal:
            # Each context is processed by a different layer,
            # good for combining modalities
            assert len(self.context_entity_bonds) == len(dim_cond_map), (
                f"Length of context entity bonds and context map sizes has to match: "
                f"{self.context_entity_bonds} != {dim_cond_map}"
            )

            layers = [
                nn.Dense(dim_cond_map[i], use_bias=True) for i in range(len(contexts))
            ]
            embeddings = [
                self.act_fn(layers[i](context)) for i, context in enumerate(contexts)
            ]
            cond_embedding = jnp.concatenate(embeddings, axis=1)
        else:
            # We can process arbitrary number of contexts, all from the same modality,
            # via a permutation-invariant deep set layer.
            sizes = [c.shape[-1] for c in contexts]
            if not len(set(sizes)) == 1:
                raise ValueError(
                    f"For embedding a set, all contexts need same length, not {sizes}"
                )
            layer = nn.Dense(dim_cond_map[0], use_bias=True)
            embeddings = [self.act_fn(layer(context)) for context in contexts]
            # Average along stacked dimension (alternatives like summing are possible)
            cond_embedding = jnp.mean(jnp.stack(embeddings), axis=0)

        z = jnp.concatenate((x, cond_embedding), axis=1)
        if self.layer_norm:
            n = nn.LayerNorm()
            z = n(z)

        for n_hidden in self.dim_hidden:
            wx = nn.Dense(n_hidden, use_bias=True)
            z = self.act_fn(wx(z))
        wx = nn.Dense(n_input, use_bias=True)

        return x + wx(z)

    def create_train_state(
        self,
        rng: jnp.ndarray,
        optimizer: optax.OptState,
        **kwargs: Any,
    ) -> PotentialTrainState:
        """Create initial `TrainState`."""
        c = jnp.ones((1, self.dim_cond))  # (n_batch, embed_dim)
        x = jnp.ones((1, self.dim_data))  # (n_batch, data_dim)
        params = self.init(rng, x=x, c=c)["params"]
        return PotentialTrainState.create(
            apply_fn=self.apply,
            params=params,
            tx=optimizer,
            potential_value_fn=self.potential_value_fn,
            potential_gradient_fn=self.potential_gradient_fn,
            **kwargs,
        )



from typing import Dict, Iterator, Optional, Tuple

import jax
import jax.numpy as jnp
import scanpy as sc
from dotmap import DotMap
from loguru import logger

from cmonge.datasets.single_loader import AbstractDataModule, DataModuleFactory
from cmonge.trainers.ae_trainer import AETrainerModule


class ConditionalDataModule:
    """Class for conditional dataloaders.

    Available modes:
    random - Conditions are randomly split into train and test, train on all data from some conditions and test on all data from others.
    homogeneous - Train on all conditions but keep out some data from every condition for testing.
    """

    datamodule_factory: Dict[str, AbstractDataModule] = DataModuleFactory

    def __init__(
        self,
        data_config: DotMap,
        condition_config: DotMap,
        ae_config: Optional[DotMap] = None,
    ):
        self.data_config = data_config
        self.name = f"cond-{data_config.name}"
        assert condition_config.mode in ["extrapolate", "homogeneous", "custom"]
        self.condition_config = condition_config
        self.ae_config = ae_config
        if self.condition_config.mode == "homogeneous":
            self.drug_condition = "homogeneous"
        else:
            self.drug_condition = data_config.drug_condition
        self.batch_size = self.data_config.batch_size
        self.key = jax.random.PRNGKey(data_config.seed)
        self.meta_loader = self.datamodule_factory[data_config.name]
        self.loaders: Dict[str, AbstractDataModule] = {}

        # Instead of loading the AnnData object from disk for every loader
        # we just pass it by reference from the "parent" class
        if self.data_config.file_path:
            self.adata = sc.read_h5ad(self.data_config.file_path)
            self.data_config.parent = self.adata
        if self.data_config.reduction:
            self.reducer()
        self.splitter()

    def sample_condition(self, split_type: str):
        """Return a random condition based on the train/valid/test split."""
        assert split_type in ["train", "valid", "test"]
        if split_type == "train":
            conditions = self.train_conditions
        elif split_type == "valid":
            conditions = self.valid_conditions
        elif split_type == "test":
            conditions = self.test_conditions

        idxs = jnp.arange(len(conditions))
        k1, self.key = jax.random.split(self.key, 2)
        idx = jax.random.choice(k1, idxs)
        return conditions[idx]

    def set_conditions(self):
        self.conditions = list(set(self.condition_config.conditions))

        self.train_conditions = []
        self.valid_conditions = []
        self.test_conditions = []
        for cond, loader in self.loaders.items():
            if len(loader.target_train_cells) > 0:
                self.train_conditions.append(cond)
            if len(loader.target_valid_cells) > 0:
                self.valid_conditions.append(cond)
            if len(loader.target_test_cells) > 0:
                self.test_conditions.append(cond)

    def setup_single_loader(self, condition: str):
        """Iitializes a loader instance for a single condition."""
        logger.info(f"Setting up datamodules for {condition}")
        self.data_config.drug_condition = condition
        loader_instance = self.meta_loader(self.data_config)
        self.loaders[condition] = loader_instance

    def decoder(self, x):
        loader = list(self.loaders.values())[0]
        return loader.decoder(x)

    @property
    def marker_idx(self):
        loader = list(self.loaders.values())[0]
        return loader.marker_idx

    def reducer(self):
        trainer = AETrainerModule(self.ae_config)
        print(self.name, self.drug_condition)
        trainer.load_model(self.name, self.drug_condition)
        self.data_config.parent_reducer = trainer

    def splitter(self):
        """Performs trian/test/valid splits based on the mode variable."""
        if self.condition_config.mode == "extrapolate":
            for condition in self.condition_config.conditions:
                if condition in self.condition_config.ood:
                    self.data_config.split = self.condition_config.ood_split or [0, 1]
                else:
                    self.data_config.split = self.condition_config.split
                self.setup_single_loader(condition)

        elif self.condition_config.mode == "homogeneous":
            self.drug_condition = "homogeneous"
            self.data_config.split = self.condition_config.split
            for condition in self.condition_config.conditions:
                self.setup_single_loader(condition)

        elif self.condition_config.mode == "custom":
            for condition, split in self.condition_config.conditions.items():
                self.data_config.split = split
                self.setup_single_loader(condition)

        self.set_conditions()

    def get_loaders_by_type(self, split_type: str):
        assert split_type in ["train", "valid", "test"]
        type_to_conditions = {
            "train": self.train_conditions,
            "valid": self.valid_conditions,
            "test": self.test_conditions,
        }

        cond_to_loaders = {
            cond: loader.get_loaders_by_type(split_type)
            for cond, loader in self.loaders.items()
            if cond in type_to_conditions[split_type]
        }
        if self.data_config.ae:

            def collapser_iter(split_type: str) -> Iterator[jnp.ndarray]:
                while len(cond_to_loaders) > 0:
                    condition = self.sample_condition(split_type)
                    loader = cond_to_loaders.get(condition, None)
                    try:
                        yield next(loader)
                    except StopIteration:
                        del cond_to_loaders[condition]
                    except TypeError:
                        pass

            loader = collapser_iter(split_type)
            return loader
        else:
            return cond_to_loaders

    def train_dataloaders(
        self,
    ) -> Dict[
        str,
        Iterator[jnp.ndarray] | Tuple[Iterator[jnp.ndarray]] | Iterator[jnp.ndarray],
    ]:
        logger.info("Setting up train dataloaders.")
        return self.get_loaders_by_type("train")

    def valid_dataloaders(
        self,
    ) -> Dict[
        str,
        Iterator[jnp.ndarray] | Tuple[Iterator[jnp.ndarray]] | Iterator[jnp.ndarray],
    ]:
        logger.info("Setting up valid dataloaders.")
        return self.get_loaders_by_type("valid")

    def test_dataloaders(
        self,
    ) -> Dict[
        str,
        Iterator[jnp.ndarray] | Tuple[Iterator[jnp.ndarray]] | Iterator[jnp.ndarray],
    ]:
        logger.info("Setting up test dataloaders.")
        return self.get_loaders_by_type("test")



from pathlib import Path
from typing import List

import jax.numpy as jnp
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from loguru import logger
from sklearn import manifold
from tqdm import tqdm

from cmonge.datasets.conditional_loader import ConditionalDataModule
from cmonge.metrics import wasserstein_distance
from cmonge.models.rdkit import rdkit_feats


class BaseEmbedding:
    def __init__(self, batch_size) -> None:
        self.batch_size = batch_size
        self.embeddings = {}


class DoseEmbedding(BaseEmbedding):
    def __init__(self, datamodule: ConditionalDataModule) -> None:
        super().__init__(datamodule.batch_size)
        self.datamodule = datamodule
        embeddings = {
            cond: [int(cond.split("-")[1]) / 10000]
            for cond in self.datamodule.conditions
        }
        self.embeddings = embeddings

    def __call__(self, condition: str):
        condition = self.embeddings[condition]
        condition_batch = jnp.asarray([condition for _ in range(self.batch_size)])
        return condition_batch


class RDKitEmbedding(BaseEmbedding):
    def __init__(
        self,
        checkpoint: bool,
        smile_path: str,
        drug_to_smile_path: str,
        name: str,
        model_dir: str,
        datamodule: ConditionalDataModule,
        dose_split: bool = True,
        drug_list: List = [
            "abexinostat",
            "belinostat",
            "dacinostat",
            "entinostat",
            "givinostat",
            "mocetinostat",
            "pracinostat",
            "tacedinaline",
            "trametinib",
        ],
    ) -> None:
        super().__init__(datamodule.batch_size)
        self.smile_path = Path(smile_path)
        self.drug_to_smile_path = Path(drug_to_smile_path)
        self.model_dir = Path(model_dir)
        self.dose_split = dose_split
        if not checkpoint:
            logger.info("Calculating RDKit embedding vectors")
            smiles_df = pd.read_csv(self.smile_path)
            smiles_list = smiles_df["smiles"].values

            # caclulate mebeddings
            n_jobs = 16
            data = Parallel(n_jobs=n_jobs)(
                delayed(rdkit_feats)(smiles)
                for smiles in tqdm(smiles_list, position=0, leave=True)
            )

            # clean data, drop nans and infs
            embedding = np.array(data)

            drug_idx, feature_idx = np.where(np.isnan(embedding))
            drug_idx_infs, feature_idx_infs = np.where(np.isinf(embedding))

            drug_idx = np.concatenate((drug_idx, drug_idx_infs))
            feature_idx = np.concatenate((feature_idx, feature_idx_infs))
            embedding[drug_idx, feature_idx] = 0

            # load smile representation of drugs
            drugs = pd.read_csv(self.drug_to_smile_path)
            drugs["drug"] = drugs["drug"].apply(lambda x: x.lower())
            drugs = drugs[drugs["drug"].isin(drug_list)][["drug", "smile"]]

            df = pd.DataFrame(
                data=embedding,
                index=smiles_list,
                columns=[f"latent_{i}" for i in range(embedding.shape[1])],
            )

            # Drop first feature from generator (RDKit2D_calculated)
            df.drop(columns=["latent_0"], inplace=True)

            # Drop columns with 0 standard deviation
            threshold = 0.01
            columns = [f"latent_{idx+1}" for idx in np.where(df.std() <= threshold)[0]]
            print(f"Deleting columns with std<={threshold}: {columns}")
            df.drop(
                columns=[f"latent_{idx+1}" for idx in np.where(df.std() <= 0.01)[0]],
                inplace=True,
            )

            # normalize and merge with drug names
            normalized_df = (df - df.mean()) / df.std()
            df = drugs.merge(normalized_df, left_on="smile", right_index=True)

            self.model_dir.mkdir(parents=True, exist_ok=True)
            model_dir = self.model_dir / name
            df.to_csv(model_dir)
        else:
            logger.info("Loading RDKit embeddings.")
            model_dir = self.model_dir / name
            df = pd.read_csv(model_dir)

        filter_col = [col for col in df if col.startswith("latent")]
        for index, row in df.iterrows():
            name = row["drug"]
            values = jnp.asarray(row[filter_col].values.astype("float"))
            self.embeddings[name] = values

    def __call__(self, condition: str):
        if self.dose_split:
            cond, dose = condition.split("-")
            condition = self.embeddings[cond]
            condition = jnp.append(condition, np.log(int(dose)))
            n_contexts = 2
        else:
            sub_conditions = condition.split("_")
            n_contexts = len(sub_conditions)
            sub_conditions = [self.embeddings[c] for c in sub_conditions]
            if len(sub_conditions) > 1:
                condition = jnp.concatenate(sub_conditions)
            else:
                condition = sub_conditions[0]
        condition_batch = jnp.asarray([condition for _ in range(self.batch_size)])
        return condition_batch, n_contexts


class ModeOfActionEmbedding(BaseEmbedding):
    def __init__(
        self,
        datamodule: ConditionalDataModule,
        checkpoint: bool,
        name: str,
        model_dir: str,
        dose_split: bool = True,
    ) -> None:
        super().__init__(datamodule.batch_size)
        self.model_dir = Path(model_dir)
        self.dose_split = dose_split
        if not checkpoint:
            labels = datamodule.train_conditions
            similarity_matrix = jnp.full(
                shape=(len(labels), len(labels)), fill_value=jnp.inf
            )

            cond_to_loaders = datamodule.train_dataloaders()
            with tqdm(total=len(labels) ** 2) as pbar:
                for i, (cond_i, loader_i) in enumerate(cond_to_loaders.items()):
                    for j, (cond_j, loader_j) in enumerate(cond_to_loaders.items()):
                        if (
                            similarity_matrix.at[i, j].get() == jnp.inf
                            and similarity_matrix.at[j, i].get() == jnp.inf
                        ):
                            target_batch_i = next(loader_i[1])
                            target_batch_j = next(loader_j[1])
                            w_dist = wasserstein_distance(
                                target_batch_i, target_batch_j
                            )
                            similarity_matrix = similarity_matrix.at[i, j].set(w_dist)
                            similarity_matrix = similarity_matrix.at[j, i].set(w_dist)
                        pbar.update(1)

            similarity = pd.DataFrame(
                data=similarity_matrix, columns=labels, index=labels
            )
            embedding, stress = manifold.smacof(
                similarity_matrix, metric=True, n_components=10
            )
            embedding_norm = (embedding - embedding.min()) / (
                embedding.max() - embedding.min()
            )
            smacof_10d = pd.DataFrame(data=embedding_norm.T, columns=labels)

            self.model_dir.mkdir(parents=True, exist_ok=True)
            model_dir = self.model_dir / name
            smacof_10d.to_csv(model_dir)
            model_dir = self.model_dir / "similarity"
            similarity.to_csv(model_dir)
        else:
            model_dir = self.model_dir / name
            smacof_10d = pd.read_csv(model_dir)
            smacof_10d = smacof_10d.drop(columns=["Unnamed: 0"])

        for index, row in smacof_10d.T.iterrows():
            values = jnp.asarray(row.values.astype("float"))
            self.embeddings[index] = values

    def __call__(self, condition: str):
        if self.dose_split:
            cond, dose = condition.split("-")
            condition = self.embeddings[condition]
            condition = jnp.append(condition, np.log(int(dose)))
            n_contexts = 2
        else:
            sub_conditions = condition.split("_")
            n_contexts = len(sub_conditions)
            enc_sub_conditions = []
            for c in sub_conditions:
                try:
                    enc_sub_conditions.append(self.embeddings[c])
                except KeyError:
                    # For the MoA embedding drugs that don't occur as
                    # single perturbation in the dataset are skipped
                    n_contexts -= 1
            if len(sub_conditions) > 1:
                condition = jnp.concatenate(enc_sub_conditions)
            else:
                condition = enc_sub_conditions[0]
        condition_batch = jnp.asarray([condition for _ in range(self.batch_size)])
        return condition_batch, n_contexts


EmbeddingFactory = {
    "rdkit": RDKitEmbedding,
    "dose": DoseEmbedding,
    "moa": ModeOfActionEmbedding,
}



import abc
from itertools import cycle
from pathlib import Path
from typing import Iterator, Optional, Tuple

import jax
import jax.numpy as jnp
import numpy as np
import scanpy as sc
from anndata import AnnData
from dotmap import DotMap
from jaxtyping import PRNGKeyArray
from loguru import logger
from sklearn.model_selection import train_test_split

from cmonge.trainers.ae_trainer import AETrainerModule
from cmonge.utils import load_config


class AbstractDataModule:
    """Abstract class for handling data loading, processing and splitting."""

    def __init__(self) -> None:
        self.adata = None
        self.encoder = lambda x: x
        self.decoder = lambda x: x
        self.reduction = None
        self.marker_idx = None
        self.ae = False

    @abc.abstractmethod
    def setup(self, *args, **kwargs) -> None:
        """Abstract method for loading, setting up the DataGenerator."""

    @abc.abstractmethod
    def train_dataloaders(self) -> Tuple[Iterator[jnp.ndarray], Iterator[jnp.ndarray]]:
        """Abstract method for creating and returning an interator from the training data."""

    @abc.abstractmethod
    def valid_dataloaders(self) -> Tuple[Iterator[jnp.ndarray], Iterator[jnp.ndarray]]:
        """Abstract method for creating and returning an iterator from the validation data."""

    @abc.abstractmethod
    def test_dataloaders(self) -> Tuple[Iterator[jnp.ndarray], Iterator[jnp.ndarray]]:
        """Abstract method for creating and returning an iterator from the test data."""

    @staticmethod
    def sampler_iter(
        array: jnp.ndarray, batch_size: int, key: PRNGKeyArray
    ) -> Iterator[jnp.ndarray]:
        """Creates an inifinite dataloader with random sampling out of a jax array."""
        while True:
            k1, key = jax.random.split(key, 2)
            yield jax.random.choice(key=k1, a=array, shape=(batch_size,))

    @staticmethod
    def batcher(array: jnp.ndarray, batch_size: int) -> jnp.ndarray:
        """Groups the data into batches of size batch size."""
        data_size = array.shape[0]
        num_batches = data_size // batch_size
        trimmed = array[: num_batches * batch_size]
        batched = trimmed.reshape((num_batches, batch_size) + array.shape[1:])
        return batched

    def batcher_iter(
        self, array: jnp.ndarray, batch_size: int
    ) -> Iterator[jnp.ndarray]:
        """Groups the data into batches and returns an iterator obejct."""
        batched = self.batcher(array, batch_size)
        for batch in batched:
            yield batch

    def cyclic_iter(self, array: jnp.ndarray, batch_size: int) -> Iterator[jnp.ndarray]:
        """Groups the data into batches and creates a cyclic iterator object."""
        batched = self.batcher(array, batch_size)
        return cycle(batched)

    def splitter(self):
        """Splits the dataset into source-target and train/valid/test buckets."""
        logger.info("Splitting dataset started.")
        source_idx = self.adata.obs[
            self.adata.obs[self.drug_col] == self.control_condition
        ].index
        self.control_adata = self.adata[source_idx, :]
        target_idx = self.adata.obs[
            self.adata.obs[self.drug_col] == self.drug_condition
        ].index
        self.target_adata = self.adata[target_idx, :]
        assert sum(self.split) == 1
        key, self.key = jax.random.split(self.key, 2)
        random_states = jax.random.randint(key, (4,), 0, 1000).tolist()
        if len(self.split) == 2:
            self.split = self.split + [0]
        if len(self.split) == 3:
            (
                self.control_train_cells,
                self.control_valid_cells,
                self.control_test_cells,
            ) = get_train_valid_test_split(
                self.control_adata.obs.index, self.split, random_states[:2]
            )
            self.target_train_cells, self.target_valid_cells, self.target_test_cells = (
                get_train_valid_test_split(
                    self.target_adata.obs.index, self.split, random_states[2:]
                )
            )
        else:
            raise ValueError("Invalid split.")
        cs1, cs2, cs3 = (
            self.control_train_cells.shape,
            self.control_valid_cells.shape,
            self.control_test_cells.shape,
        )

        ts1, ts2, ts3 = (
            self.target_train_cells.shape,
            self.target_valid_cells.shape,
            self.target_test_cells.shape,
        )

        logger.info(
            f"Target dataset number of cells - train: {ts1}, valid: {ts2}, test: {ts3}."
        )
        logger.info(
            f"Control dataset number of cells - train: {cs1}, valid: {cs2}, test: {cs3}."
        )
        logger.info("Splitting finished.")

    def get_loaders_by_type(
        self, split_type: str, batch_size: Optional[int] = None
    ) -> Tuple[Iterator[jnp.ndarray], Iterator[jnp.ndarray]]:
        """Convert adata object into control and target iterators, subset based on the split type (train/valid/test)."""
        if split_type == "train":
            control_cells = self.control_train_cells
            target_cells = self.target_train_cells
        elif split_type == "valid":
            control_cells = self.control_valid_cells
            target_cells = self.target_valid_cells
        elif split_type == "test":
            control_cells = self.control_test_cells
            target_cells = self.target_test_cells
        else:
            raise ValueError("Invalid split, split_type should be train/valid/test.")

        control = self.control_adata[control_cells, :].X
        target = self.target_adata[target_cells, :].X

        if self.ae:
            # Loader for AutoEncoder model
            loaders = self.get_ae_iter(control, target)

        else:
            # Loader for OT models
            if self.reduction:
                control = self.encoder(control)
                target = self.encoder(target)

            k1, self.key = jax.random.split(self.key, 2)
            k2, self.key = jax.random.split(self.key, 2)

            if batch_size is None:
                batch_size = self.batch_size
            loaders = self.sampler_iter(control, batch_size, k1), self.sampler_iter(
                target, batch_size, k2
            )
        return loaders

    def get_ae_iter(
        self, control: jnp.ndarray, target: jnp.ndarray
    ) -> Iterator[jnp.ndarray]:
        x = np.vstack((control, target))
        k1, self.key = jax.random.split(self.key, 2)
        x = jax.random.permutation(k1, x)
        loaders = self.batcher_iter(x, self.batch_size)
        return loaders


def get_train_valid_test_split(x: jnp.ndarray, split: list[int], seeds=[0, 1]):
    assert len(split) == 3
    assert sum(split) == 1
    train, valid, test = split
    empty = jnp.empty(0)
    if train == 0:
        x_train = empty
        x_valid_test = x
    elif train == 1:
        x_train = x
        x_valid_test = empty
    else:
        x_train, x_valid_test = train_test_split(
            x, random_state=seeds[0], test_size=(1 - train), shuffle=True
        )

    if valid == 0:
        x_valid = empty
        x_test = x_valid_test
    elif train + valid == 1:
        x_valid = x_valid_test
        x_test = empty
    else:
        x_valid, x_test = train_test_split(
            x_valid_test,
            random_state=seeds[1],
            test_size=(test / (1 - train)),
            shuffle=True,
        )
    return x_train, x_valid, x_test


class SciPlexModule(AbstractDataModule):
    def __init__(self, config: DotMap):
        super().__init__()
        self.setup(**config)

    def setup(
        self,
        name: str,
        file_path: Path,
        batch_size: int,
        split: list[float],
        drug_col: str,
        drug_condition: str,
        control_condition: str,
        seed: str,
        ae: bool = False,
        ae_config_path: Optional[Path] = None,
        reduction: Optional[str] = None,
        parent: Optional[AnnData] = None,
        parent_reducer: Optional[str] = None,
    ):
        self.name = name
        self.file_path = file_path
        self.batch_size = batch_size
        self.split = split
        self.drug_col = drug_col
        self.drug_condition = drug_condition
        self.control_condition = control_condition
        self.ae = ae
        self.ae_config = load_config(ae_config_path) if ae_config_path else None
        self.reduction = reduction
        self.parent = parent
        self.parent_reducer = parent_reducer
        self.seed = seed
        self.drug = self.drug_condition.split("-")[0]
        self.key = jax.random.PRNGKey(self.seed)

        self.loader()
        self.preprocesser()
        self.splitter()
        self.reducer()

    def loader(self) -> None:
        if self.parent:
            self.adata = self.parent
        else:
            self.adata = sc.read_h5ad(self.file_path)
        # save marker genes for evaluation
        try:
            genes = self.adata.uns["rank_genes_groups"]["names"][self.drug]
            self.marker_genes = [
                gene for gene in genes if gene in self.adata.var.index
            ][:50]
            self.gene_idx_to_enum = {
                idx: enum for (enum, idx) in enumerate(self.adata.var.index)
            }
            self.marker_idx = [
                self.gene_idx_to_enum[gene] for gene in self.marker_genes
            ]
            logger.info(
                f"{len(self.marker_idx)} marker genes are saved for evaluation."
            )
        except ValueError:
            print("Make sure the h5ad file_path points to the sciplex dataset.")

    def preprocesser(self) -> None:
        """Normalizes and log transofrms the data."""
        # sc.pp.log1p(self.adata)
        # sc.pp.normalize_total(self.adata)
        if not isinstance(self.adata.X, np.ndarray):
            self.adata.X = jnp.asarray(self.adata.X.todense())

    def reducer(self):
        """Sets up dimensionality reduction, either with PCA, AE or identity."""
        if self.reduction == "pca":
            self.pca_means = self.adata.X.mean(axis=0)
            self.encoder = lambda x: (x - self.pca_means) @ self.adata.varm["PCs"]
            self.decoder = lambda x: x @ self.adata.varm["PCs"].T + self.pca_means
        elif self.reduction == "ae":
            if self.parent_reducer:
                trainer = self.parent_reducer
            else:
                trainer = AETrainerModule(self.ae_config)
                trainer.load_model(self.name, self.drug_condition)
            model = trainer.model.bind({"params": trainer.state.params})
            self.encoder = lambda x: model.encoder(x)
            self.decoder = lambda x: model.decoder(x)
        else:
            self.encoder = lambda x: x
            self.decoder = lambda x: x

    def train_dataloaders(self) -> Tuple[Iterator[jnp.ndarray], Iterator[jnp.ndarray]]:
        """Convert training dataset into infinite iterators."""
        train_loaders = self.get_loaders_by_type("train")
        return train_loaders

    def valid_dataloaders(self) -> Tuple[Iterator[jnp.ndarray], Iterator[jnp.ndarray]]:
        """Convert test dataset into infinite iterators."""
        valid_loaders = self.get_loaders_by_type("valid")
        return valid_loaders

    def test_dataloaders(self) -> Tuple[Iterator[jnp.ndarray], Iterator[jnp.ndarray]]:
        """Convert test dataset into a batch iterator."""
        test_loaders = self.get_loaders_by_type("test")
        return test_loaders


class FourIModule(AbstractDataModule):
    def __init__(self, config: DotMap) -> None:
        super().__init__()
        self.setup(**config)

    def loader(self) -> None:
        if self.parent:
            self.adata = self.parent
        else:
            self.adata = sc.read_h5ad(self.file_path)

        with open(self.features_path) as f:
            features = f.readlines()
        self.features = [feature.rstrip() for feature in features]

        with open(self.drugs_path) as f:
            drugs = f.readlines()
        self.drugs = [drug.rstrip() for drug in drugs]

    def preprocesser(self) -> None:
        self.adata = self.adata[:, self.features]
        self.adata.X = jnp.asarray(self.adata.X)

    def setup(
        self,
        name: str,
        file_path: Path,
        drugs_path: Path,
        features: Path,
        split: list[float],
        batch_size: int,
        drug_col: str,
        drug_condition: str,
        control_condition: str,
        ae: bool,
        seed: int,
        split_dose: bool = False,
        parent: Optional[AnnData] = None,
        reduction: Optional[str] = None,  # For compatability
        parent_reducer: Optional[str] = None,  # For compatability
    ) -> None:
        self.name = name
        self.file_path = file_path
        self.split = split
        self.batch_size = batch_size
        self.features_path = features
        self.drugs_path = drugs_path
        self.drug_col = drug_col
        self.drug_condition = drug_condition
        self.control_condition = control_condition
        self.ae = ae
        self.key = jax.random.PRNGKey(seed)
        self.parent = parent
        self.reduction = reduction
        self.parent_reducer = parent_reducer
        self.loader()
        self.preprocesser()
        self.splitter()

    def train_dataloaders(self) -> Tuple[Iterator[jnp.ndarray], Iterator[jnp.ndarray]]:
        train_loaders = self.get_loaders_by_type("train")
        return train_loaders

    def valid_dataloaders(self) -> Tuple[Iterator[jnp.ndarray], Iterator[jnp.ndarray]]:
        valid_loaders = self.get_loaders_by_type("valid")
        return valid_loaders

    def test_dataloaders(
        self, batch_size: Optional[int] = None
    ) -> Tuple[Iterator[jnp.ndarray], Iterator[jnp.ndarray]]:
        test_loaders = self.get_loaders_by_type("test", batch_size)
        return test_loaders

    # For compatability with the ConditionalDataloader
    def reducer(self):
        """Sets up dimensionality reduction, either with PCA, AE or identity."""
        if self.reduction == "pca":
            self.pca_means = self.adata.X.mean(axis=0)
            self.encoder = lambda x: (x - self.pca_means) @ self.adata.varm["PCs"]
            self.decoder = lambda x: x @ self.adata.varm["PCs"].T + self.pca_means
        elif self.reduction == "ae":
            if self.parent_reducer:
                trainer = self.parent_reducer
            else:
                trainer = AETrainerModule(self.ae_config)
                trainer.load_model(self.name, self.drug_condition)
            model = trainer.model.bind({"params": trainer.state.params})
            self.encoder = lambda x: model.encoder(x)
            self.decoder = lambda x: model.decoder(x)
        else:
            self.encoder = lambda x: x
            self.decoder = lambda x: x


DataModuleFactory = {
    "4i": FourIModule,
    "sciplex": SciPlexModule,
}  # , "synthetic": SyntheticDosageModule}



from cmonge.datasets.conditional_loader import ConditionalDataModule
from cmonge.datasets.single_loader import SciPlexModule


class TestConditionalDataModule:
    def test_homogeneous_splitter(self, synthetic_config):
        config = synthetic_config.copy()
        config.condition.mode = "homogeneous"
        config.condition.conditions = ["givinostat-10", "givinostat-100"]
        config.condition.split = [0.6, 0.2, 0.2]
        module = ConditionalDataModule(config.data, config.condition, config.ae)

        module.splitter()

        assert len(module.loaders) == 2
        assert isinstance(module.loaders["givinostat-10"], SciPlexModule)

        assert len(module.loaders["givinostat-10"].control_train_cells) == 55
        assert len(module.loaders["givinostat-10"].control_valid_cells) == 18
        assert len(module.loaders["givinostat-10"].control_test_cells) == 19

    def test_extrapolate_splitter(self, synthetic_config):
        config = synthetic_config.copy()
        config.condition.mode = "extrapolate"
        config.condition.conditions = ["givinostat-10", "givinostat-100"]
        config.condition.split = [0.8, 0.2]
        config.condition.ood = ["givinostat-100"]
        config.condition.ood_split = [0, 1]
        module = ConditionalDataModule(config.data, config.condition, config.ae)

        assert len(module.loaders) == 2
        assert isinstance(module.loaders["givinostat-100"], SciPlexModule)
        train_condition = module.train_conditions[0]
        assert len(module.loaders[train_condition].control_train_cells) == 73
        assert len(module.loaders[train_condition].control_valid_cells) == 19
        assert len(module.loaders[train_condition].control_test_cells) == 0

        valid_condition = module.valid_conditions[0]
        assert len(module.loaders[valid_condition].control_train_cells) == 73
        assert len(module.loaders[valid_condition].control_valid_cells) == 19
        assert len(module.loaders[valid_condition].control_test_cells) == 0

        assert len(module.train_conditions) == 1
        assert len(module.valid_conditions) == 2

    def test_sample_condition(self, synthetic_config):
        config = synthetic_config.copy()
        config.condition.mode = "extrapolate"
        config.condition.conditions = ["givinostat-10", "givinostat-100"]
        config.condition.split = [1, 0]
        config.condition.ood = ["givinostat-100"]
        module = ConditionalDataModule(config.data, config.condition, config.ae)

        train_cond = module.sample_condition("train")
        valid_cond = module.sample_condition("valid")

        assert train_cond in "givinostat-10"
        assert valid_cond == "givinostat-100"



from loguru import logger
from rdkit import Chem
from rdkit.Chem import Descriptors

RDKIT_PROPS = [
    "BalabanJ",
    "BertzCT",
    "Chi0",
    "Chi0n",
    "Chi0v",
    "Chi1",
    "Chi1n",
    "Chi1v",
    "Chi2n",
    "Chi2v",
    "Chi3n",
    "Chi3v",
    "Chi4n",
    "Chi4v",
    "EState_VSA1",
    "EState_VSA10",
    "EState_VSA11",
    "EState_VSA2",
    "EState_VSA3",
    "EState_VSA4",
    "EState_VSA5",
    "EState_VSA6",
    "EState_VSA7",
    "EState_VSA8",
    "EState_VSA9",
    "ExactMolWt",
    "FpDensityMorgan1",
    "FpDensityMorgan2",
    "FpDensityMorgan3",
    "FractionCSP3",
    "HallKierAlpha",
    "HeavyAtomCount",
    "HeavyAtomMolWt",
    "Ipc",
    "Kappa1",
    "Kappa2",
    "Kappa3",
    "LabuteASA",
    "MaxAbsEStateIndex",
    "MaxAbsPartialCharge",
    "MaxEStateIndex",
    "MaxPartialCharge",
    "MinAbsEStateIndex",
    "MinAbsPartialCharge",
    "MinEStateIndex",
    "MinPartialCharge",
    "MolLogP",
    "MolMR",
    "MolWt",
    "NHOHCount",
    "NOCount",
    "NumAliphaticCarbocycles",
    "NumAliphaticHeterocycles",
    "NumAliphaticRings",
    "NumAromaticCarbocycles",
    "NumAromaticHeterocycles",
    "NumAromaticRings",
    "NumHAcceptors",
    "NumHDonors",
    "NumHeteroatoms",
    "NumRadicalElectrons",
    "NumRotatableBonds",
    "NumSaturatedCarbocycles",
    "NumSaturatedHeterocycles",
    "NumSaturatedRings",
    "NumValenceElectrons",
    "PEOE_VSA1",
    "PEOE_VSA10",
    "PEOE_VSA11",
    "PEOE_VSA12",
    "PEOE_VSA13",
    "PEOE_VSA14",
    "PEOE_VSA2",
    "PEOE_VSA3",
    "PEOE_VSA4",
    "PEOE_VSA5",
    "PEOE_VSA6",
    "PEOE_VSA7",
    "PEOE_VSA8",
    "PEOE_VSA9",
    "RingCount",
    "SMR_VSA1",
    "SMR_VSA10",
    "SMR_VSA2",
    "SMR_VSA3",
    "SMR_VSA4",
    "SMR_VSA5",
    "SMR_VSA6",
    "SMR_VSA7",
    "SMR_VSA8",
    "SMR_VSA9",
    "SlogP_VSA1",
    "SlogP_VSA10",
    "SlogP_VSA11",
    "SlogP_VSA12",
    "SlogP_VSA2",
    "SlogP_VSA3",
    "SlogP_VSA4",
    "SlogP_VSA5",
    "SlogP_VSA6",
    "SlogP_VSA7",
    "SlogP_VSA8",
    "SlogP_VSA9",
    "TPSA",
    "VSA_EState1",
    "VSA_EState10",
    "VSA_EState2",
    "VSA_EState3",
    "VSA_EState4",
    "VSA_EState5",
    "VSA_EState6",
    "VSA_EState7",
    "VSA_EState8",
    "VSA_EState9",
    "fr_Al_COO",
    "fr_Al_OH",
    "fr_Al_OH_noTert",
    "fr_ArN",
    "fr_Ar_COO",
    "fr_Ar_N",
    "fr_Ar_NH",
    "fr_Ar_OH",
    "fr_COO",
    "fr_COO2",
    "fr_C_O",
    "fr_C_O_noCOO",
    "fr_C_S",
    "fr_HOCCN",
    "fr_Imine",
    "fr_NH0",
    "fr_NH1",
    "fr_NH2",
    "fr_N_O",
    "fr_Ndealkylation1",
    "fr_Ndealkylation2",
    "fr_Nhpyrrole",
    "fr_SH",
    "fr_aldehyde",
    "fr_alkyl_carbamate",
    "fr_alkyl_halide",
    "fr_allylic_oxid",
    "fr_amide",
    "fr_amidine",
    "fr_aniline",
    "fr_aryl_methyl",
    "fr_azide",
    "fr_azo",
    "fr_barbitur",
    "fr_benzene",
    "fr_benzodiazepine",
    "fr_bicyclic",
    "fr_diazo",
    "fr_dihydropyridine",
    "fr_epoxide",
    "fr_ester",
    "fr_ether",
    "fr_furan",
    "fr_guanido",
    "fr_halogen",
    "fr_hdrzine",
    "fr_hdrzone",
    "fr_imidazole",
    "fr_imide",
    "fr_isocyan",
    "fr_isothiocyan",
    "fr_ketone",
    "fr_ketone_Topliss",
    "fr_lactam",
    "fr_lactone",
    "fr_methoxy",
    "fr_morpholine",
    "fr_nitrile",
    "fr_nitro",
    "fr_nitro_arom",
    "fr_nitro_arom_nonortho",
    "fr_nitroso",
    "fr_oxazole",
    "fr_oxime",
    "fr_para_hydroxylation",
    "fr_phenol",
    "fr_phenol_noOrthoHbond",
    "fr_phos_acid",
    "fr_phos_ester",
    "fr_piperdine",
    "fr_piperzine",
    "fr_priamide",
    "fr_prisulfonamd",
    "fr_pyridine",
    "fr_quatN",
    "fr_sulfide",
    "fr_sulfonamd",
    "fr_sulfone",
    "fr_term_acetylene",
    "fr_tetrazole",
    "fr_thiazole",
    "fr_thiocyan",
    "fr_thiophene",
    "fr_unbrch_alkane",
    "fr_urea",
    "qed",
]
FEAT_FNS = {name: func for name, func in Descriptors.descList}


def safely_compute_feature(name: str, mol):
    try:
        return FEAT_FNS[name](mol)
    except Exception:
        logger.exception(f"Failed to apply {name} on {Chem.MolToSmiles(mol)}")


def rdkit_feats(smiles):
    """
    Calculate RDKit 2D descriptors for a given SMILES string.

    Returns a numpy array of descriptor values or None if the SMILES string is invalid.
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return None
    values = [safely_compute_feature(name, mol) for name in RDKIT_PROPS]
    return [True] + values



from typing import Any, Dict

import jax.numpy as jnp
from loguru import logger

from cmonge.metrics import (
    average_r2,
    compute_scalar_mmd,
    drug_signature,
    eucledian_monge_gap,
    sinkhorn_div,
    wasserstein_distance,
)


def init_logger_dict(metrics: Dict[str, Any], drug) -> Dict[str, Any]:
    metrics["point_clouds"] = {"source": [], "target": [], "transport": []}
    metrics["drug"] = drug
    metrics["wasserstein"] = []
    metrics["mmd"] = []
    metrics["sinkhorn_div"] = []
    metrics["monge_gap"] = []
    metrics["drug_signature"] = []
    metrics["r2"] = []
    metrics["mean_statistics"] = {}
    return metrics


def log_metrics(metrics, target, transport):
    w_dist = wasserstein_distance(target, transport)
    mmd_dist = compute_scalar_mmd(target, transport)
    sh_div = sinkhorn_div(target, transport)
    monge_gap = eucledian_monge_gap(target, transport)
    ds = drug_signature(target, transport)
    r2 = average_r2(target, transport)

    metrics["wasserstein"].append(w_dist)
    metrics["mmd"].append(mmd_dist)
    metrics["sinkhorn_div"].append(sh_div)
    metrics["monge_gap"].append(monge_gap)
    metrics["drug_signature"].append(ds)
    metrics["r2"].append(r2)


def log_point_clouds(metrics, source, target, transport):
    metrics["point_clouds"]["source"].append(source)
    metrics["point_clouds"]["target"].append(target)
    metrics["point_clouds"]["transport"].append(transport)


def log_mean_metrics(metrics):
    metrics["mean_statistics"]["mean_wasserstein"] = float(
        sum(metrics["wasserstein"]) / len(metrics["wasserstein"])
    )
    metrics["mean_statistics"]["mean_mmd"] = float(
        sum(metrics["mmd"]) / len(metrics["mmd"])
    )
    metrics["mean_statistics"]["mean_sinkhorn div"] = float(
        sum(metrics["sinkhorn_div"]) / len(metrics["sinkhorn_div"])
    )
    metrics["mean_statistics"]["mean_monge_gap"] = float(
        sum(metrics["monge_gap"]) / len(metrics["monge_gap"])
    )
    metrics["mean_statistics"]["mean_drug_signature"] = float(
        sum(metrics["drug_signature"]) / len(metrics["drug_signature"])
    )
    metrics["mean_statistics"]["mean_r2"] = float(
        sum(metrics["r2"]) / len(metrics["r2"])
    )
    logger.info(metrics["mean_statistics"])


def get_single_loaders_for_eval(datamodule, split):
    if split == "valid":
        loader_source, loader_target = datamodule.valid_dataloaders()
    else:
        loader_source, loader_target = datamodule.test_dataloaders()
    return loader_source, loader_target


def get_conditional_loaders_for_eval(datamodule):
    condition = datamodule.sample_condition("valid")
    valid_condition_to_loaders = datamodule.valid_dataloaders()
    loader_source, loader_target = valid_condition_to_loaders[condition]
    condition = datamodule.embeddings[condition]
    cond = jnp.asarray([[condition] for _ in range(datamodule.batch_size)])
    return loader_source, loader_target, condition, cond



import abc
import warnings
from functools import partial
from pathlib import Path
from typing import Any, Dict, Optional, Type, TypeVar

import jax.numpy as jnp
import optax
from dotmap import DotMap
from flax import linen as nn
from flax.training.orbax_utils import save_args_from_target
from jax.lib import xla_bridge
from loguru import logger
from orbax.checkpoint import PyTreeCheckpointer
from ott.neural.methods.monge_gap import MongeGapEstimator
from ott.neural.methods.neuraldual import W2NeuralDual
from ott.neural.networks.icnn import ICNN
from ott.neural.networks.potentials import PotentialMLP

from cmonge.datasets.single_loader import AbstractDataModule
from cmonge.evaluate import (
    init_logger_dict,
    log_mean_metrics,
    log_metrics,
    log_point_clouds,
)
from cmonge.metrics import fitting_loss, regularizer
from cmonge.utils import create_or_update_logfile, optim_factory

T = TypeVar("T", bound="AbstractTrainer")


class AbstractTrainer(abc.ABC):
    """Abstract class for neural OT traininig."""

    def __init__(self, jobid: int, logger_path: Path) -> None:
        self.jobid = jobid
        self.logger_path = logger_path
        self.metrics: Dict[str, Any] = {}
        self.metrics["jobid"] = jobid
        device = xla_bridge.get_backend().platform
        logger.info(f"JAX uses {device} for trianing.")

    @abc.abstractmethod
    def setup(
        self,
        *args,
        **kwargs,
    ):
        """Abstract method for setting up the model."""
        pass

    @abc.abstractmethod
    def train(self, datamodule: AbstractDataModule) -> None:
        """Abstract method for training loop."""
        pass

    @abc.abstractmethod
    def transport(self, source: jnp.ndarray) -> jnp.ndarray:
        """Abstract method for transporting a batch of data."""
        pass

    @property
    @abc.abstractmethod
    def model(self) -> nn.Module:
        """Abstract method that returns the attribute to be checkpointed."""
        pass

    @model.setter
    @abc.abstractmethod
    def model(self, value: nn.Module):
        """Abstract property to set the checkpointed attribute."""
        pass

    def save_checkpoint(
        self, path: Optional[Path] = None, config: Optional[DotMap] = None
    ) -> None:
        """Abstract method for saving model parameters to a pickle file.

        Args:
            path: Path where the checkpoint should be saved. Defaults to None in which case
                it is retrieved from config.
            config: The model training configuration with a `checkpointing_path` field.
                Defaults to None.
                NOTE: If `config` and `path` are both not None, `path` takes preference.
        """
        if path is None and config is None:
            raise ValueError(
                "Checkpoint cannot be saved. Provide a checkpoint save path either directly or through the config."
            )
        elif path is None:
            path = config.checkpointing_path
        try:
            checkpointer = PyTreeCheckpointer()
            save_args = save_args_from_target(self.model)
            checkpointer.save(path, self.model, save_args=save_args, force=True)
        except Exception as e:
            raise Exception(f"Error in saving checkpoint to {path}: {e}")

    @classmethod
    def load_checkpoint(
        cls: Type[T],
        jobid: int,
        logger_path: Path,
        config: DotMap,
        ckpt_path: Path = None,
        *args,
        **kwargs,
    ) -> T:
        """
        Loading a model from a checkpoint

        Args:
            cls: Class object to be created.
            jobid: ID of the job.
            logger_path: Path where the logging files are stored.
            config: Model training configuration.
            ckpt_path: Optional path from where checkpoint is restored.
                Defaults to None, in that case inferred from config.

        Returns:
            Class object with restored weights.
        """
        try:
            out_class = cls(
                jobid=jobid, logger_path=logger_path, config=config, *args, **kwargs
            )

            if ckpt_path is None:
                if len(config.checkpointing_path) > 0:
                    ckpt_path = config.checkpointing_path
                else:
                    logger.error(
                        "Provide checkpointing path either directly or through the model config"
                    )
            checkpointer = PyTreeCheckpointer()
            out_class.model = checkpointer.restore(ckpt_path, item=out_class.model)

            logger.info("Loaded model from checkpoint")
            return out_class
        except Exception as e:
            raise Exception(
                f"Failed to load checkpoin from {ckpt_path}: {e}\nAre you sure checkpoint was saved and correct path is provided?"
            )

    def evaluate(
        self,
        datamodule: AbstractDataModule,
        identity: bool = False,
        valid: bool = False,
        n_samples: int = 9,
        log_transport: bool = False,
    ) -> None:
        """Evaluate a trained model on a validation set and save the metrics to a json file."""
        logger.info(f"Evaluation started on {datamodule.drug_condition}.")
        init_logger_dict(self.metrics, datamodule.drug_condition)
        if valid:
            loader_source, loader_target = datamodule.valid_dataloaders()
            self.metrics["eval"] = "valid"
        else:
            loader_source, loader_target = datamodule.test_dataloaders()
            self.metrics["eval"] = "test"
        for enum, (source, target) in enumerate(zip(loader_source, loader_target)):
            if not identity:
                transport = self.transport(source)

            else:
                transport = source

            transport = datamodule.decoder(transport)
            target = datamodule.decoder(target)
            if log_transport:
                log_point_clouds(self.metrics, source, target, transport)
            if datamodule.marker_idx:
                target = target[:, datamodule.marker_idx]
                transport = transport[:, datamodule.marker_idx]

            log_metrics(self.metrics, target, transport)
            if enum > n_samples:
                break

        log_mean_metrics(self.metrics)
        create_or_update_logfile(self.logger_path, self.metrics)


class MongeGapTrainer(AbstractTrainer):
    """Wrapper class for Monge Gap training."""

    def __init__(self, jobid: int, logger_path: Path, config: DotMap) -> None:
        super().__init__(jobid, logger_path)
        self.setup(**config)

    def setup(
        self,
        method: str,
        dim_hidden: list[int],
        num_genes: int,
        num_train_iters: int,
        fitting_loss: Dict[str, Any],
        regularizer: Dict[str, Any],
        optim: Dict[str, Any],
        checkpointing_path: Optional[str] = None,  # For compatibility with base class
    ) -> None:
        """Initializes models and optimizers."""
        self.metrics["params"] = {
            "num_genes": num_genes,
            "num_train_iters": num_train_iters,
            "fitting_loss": fitting_loss.name,
            "regularizer": regularizer.name,
            "dim_hidden": dim_hidden,
            "lr": optim.lr,
            "method": method,
        }

        # setup loss function and regularizer
        fitting_loss_fn = loss_factory[fitting_loss.name]
        regularizer_fn = regularizer_factory[regularizer.name]
        fitting_loss = partial(fitting_loss_fn, **fitting_loss.kwargs)
        regularizer = partial(regularizer_fn, **regularizer.kwargs)

        # setup neural network model
        model = PotentialMLP(dim_hidden=dim_hidden, is_potential=False, act_fn=nn.gelu)

        # setup optimizer and scheduler
        opt_fn = optim_factory[optim.name]
        lr_scheduler = optax.cosine_decay_schedule(
            init_value=optim.lr, decay_steps=num_train_iters, alpha=1e-2
        )
        optimizer = opt_fn(learning_rate=lr_scheduler, **optim.kwargs)

        # setup ott-jax solver
        self.solver = MongeGapEstimator(
            num_genes,
            fitting_loss=fitting_loss,
            regularizer=regularizer,
            model=model,
            optimizer=optimizer,
            regularizer_strength=1,
            num_train_iters=num_train_iters,
            logging=True,
            valid_freq=100,
        )

    def train(self, datamodule: AbstractDataModule) -> None:
        """Trains a Monge Gap estimator."""
        logger.info("Training started")
        train_loader_source, train_loader_target = datamodule.train_dataloaders()
        valid_loader_source, valid_loader_target = datamodule.valid_dataloaders()
        self.solver.state_neural_net, logs = self.solver.train_map_estimator(
            trainloader_source=train_loader_source,
            trainloader_target=train_loader_target,
            validloader_source=valid_loader_source,
            validloader_target=valid_loader_target,
        )
        self.metrics["ottlogs"] = logs
        logger.info("Training finished.")

    def transport(self, source: jnp.ndarray) -> jnp.ndarray:
        """Transports a batch of data using the learned model."""
        return self.solver.state_neural_net.apply_fn(
            {"params": self.solver.state_neural_net.params}, source
        )

    @property
    def model(self) -> nn.Module:
        return self.solver.state_neural_net

    @model.setter
    def model(self, value: nn.Module):
        """Setter for the model to be checkpointed."""
        self.solver.state_neural_net = value


class MongeMapTrainer(MongeGapTrainer):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "MongeMapTrainer is deprecated and will be removed in a future release. "
            "Please use MongeGapTrainer instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)


class NeuralDualTrainer(AbstractTrainer):
    """Wrapper class for neural dual training."""

    def __init__(self, jobid: int, logger_path: Path, config: DotMap) -> None:
        super().__init__(jobid, logger_path)
        self.setup(**config)

    def setup(
        self,
        dim_hidden: list[int],
        lr: float,
        num_genes: int,
        num_train_iters: int,
        num_inner_iters: int,
        samples_source: jnp.ndarray,
        samples_target: jnp.ndarray,
        **kwargs,
    ) -> None:
        """Initializes models and optimizers."""
        self.metrics["params"] = {
            "num_genes": num_genes,
            "num_training_iters": num_train_iters,
            "num_inner_iters": num_inner_iters,
            "dim_hidden": dim_hidden,
            "lr": lr,
            "method": "dual",
        }
        neural_f = ICNN(
            dim_data=num_genes,
            dim_hidden=dim_hidden,
            gaussian_map_samples=(samples_source, samples_target),
        )
        neural_g = PotentialMLP(dim_hidden=dim_hidden)

        lr_schedule = optax.cosine_decay_schedule(
            init_value=lr, decay_steps=num_train_iters, alpha=1e-2
        )
        optimizer_f = optax.adamw(learning_rate=lr_schedule, b1=0.5, b2=0.5)
        optimizer_g = optax.adamw(learning_rate=lr_schedule, b1=0.9, b2=0.999)

        self.neural_dual_solver = W2NeuralDual(
            num_genes,
            neural_f,
            neural_g,
            optimizer_f,
            optimizer_g,
            num_train_iters=num_train_iters,
            logging=True,
            num_inner_iters=num_inner_iters,
            valid_freq=100,
            log_freq=100,
            amortization_loss="objective",
        )

    def train(self, datamodule: AbstractDataModule) -> None:
        """Trains Dual Estimator."""
        logger.info("Training started")
        train_loader_source, train_loader_target = datamodule.train_dataloaders()
        valid_loader_source, valid_loader_target = datamodule.valid_dataloaders()
        result = self.neural_dual_solver(
            train_loader_source,
            train_loader_target,
            valid_loader_source,
            valid_loader_target,
        )
        if isinstance(result, tuple):
            pot, logs = result
            self.potentials = pot
            self.metrics["ottlogs"] = logs
            logger.info("Training finished")

    def transport(self, source: jnp.ndarray) -> jnp.ndarray:
        """Transports a batch of data using the Brenier formula."""
        return self.potentials.transport(source)

    @property
    def model(self) -> nn.Module:
        return self.neural_dual_solver

    @model.setter
    def model(self, value: nn.Module):
        """Setter for the model to be checkpointed."""
        self.neural_dual_solver = value


loss_factory = {"sinkhorn": fitting_loss}
regularizer_factory = {"monge": regularizer}



import collections
import functools
from functools import partial
from pathlib import Path
from typing import Callable, Dict, Iterator, Optional, Tuple, Union

import flax.linen as nn
import jax
import jax.numpy as jnp
import optax
from dotmap import DotMap
from flax.core import frozen_dict
from flax.training import train_state
from flax.training.orbax_utils import save_args_from_target
from jax.tree_util import tree_map
from loguru import logger
from orbax.checkpoint import PyTreeCheckpointer

from cmonge.datasets.conditional_loader import ConditionalDataModule
from cmonge.evaluate import init_logger_dict, log_mean_metrics, log_metrics
from cmonge.models.embedding import BaseEmbedding, EmbeddingFactory
from cmonge.models.nn import ConditionalPerturbationNetwork
from cmonge.trainers.ot_trainer import (
    AbstractTrainer,
    loss_factory,
    regularizer_factory,
)
from cmonge.utils import create_or_update_logfile, optim_factory


class ConditionalMongeTrainer(AbstractTrainer):
    embedding_factory: Dict[str, BaseEmbedding] = EmbeddingFactory

    def __init__(
        self,
        jobid: int,
        logger_path: Path,
        config: DotMap,
        datamodule: ConditionalDataModule,
    ) -> None:
        super().__init__(jobid, logger_path)
        self.config = config
        self.datamodule = datamodule

        self.key = jax.random.PRNGKey(self.config.seed)
        self.regularizer_strength = 1
        self.num_train_iters = self.config.num_train_iters
        self.grad_acc_steps = self.config.optim.get("grad_acc_steps", 1)
        self.setup(datamodule=datamodule)

    def setup(self, datamodule: ConditionalDataModule):
        """
        Setup function has to be overriden since it is abstract in base.

        Args:
            datamodule (ConditionalDataModule): The datamodule to be trained on.
        """
        # setup loss function and regularizer
        fitting_loss_fn = loss_factory[self.config.fitting_loss.name]
        regularizer_fn = regularizer_factory[self.config.regularizer.name]
        self.fitting_loss = partial(fitting_loss_fn, **self.config.fitting_loss.kwargs)
        self.regularizer = partial(regularizer_fn, **self.config.regularizer.kwargs)

        # setup optimizer and scheduler
        opt_fn = optim_factory[self.config.optim.name]
        lr_scheduler = self.config.get("lr_scheduler", DotMap({"name": "cosine"}))
        if lr_scheduler.name.lower() == "cosine":
            lr_scheduler = optax.cosine_decay_schedule(
                init_value=self.config.optim.lr,
                decay_steps=self.num_train_iters,
                alpha=1e-2,
            )
        elif lr_scheduler.name.lower() == "linear":
            lr_scheduler = optax.linear_onecycle_schedule(
                transition_steps=self.num_train_iters,
                peak_value=self.config.optim.lr,
                **lr_scheduler.kwargs,
            )
        optimizer = opt_fn(learning_rate=lr_scheduler, **self.config.optim.kwargs)

        self.neural_net = ConditionalPerturbationNetwork(
            **self.config.mlp
        )  # TODO: create embedding and model factory

        embed_module = self.embedding_factory[self.config.embedding.name]
        self.embedding_module = embed_module(
            datamodule=datamodule, **self.config.embedding
        )

        self.step_fn = self._get_step_fn()
        self.key, rng = jax.random.split(self.key, 2)
        self.state_neural_net = self.neural_net.create_train_state(rng, optimizer)

    def generate_batch(
        self,
        datamodule: ConditionalDataModule,
        split_type: str,
    ) -> Dict[str, jnp.ndarray]:
        """Generate a batch of condition and samples."""
        condition = datamodule.sample_condition(split_type)
        loaders = datamodule.loaders[condition]
        loader_source, loader_target = loaders.get_loaders_by_type(split_type)
        embeddings, n_contexts = self.embedding_module(condition=condition)
        return (
            {
                "source": next(loader_source),
                "target": next(loader_target),
                "condition": embeddings,
                "num_contexts": n_contexts,
            },
            condition,
        )

    def update_logs(self, current_logs, logs, tbar):
        # store and print metrics if logging step

        for log_key in current_logs:
            for metric_key in current_logs[log_key]:
                logs[log_key][metric_key].append(current_logs[log_key][metric_key])

        # update the tqdm bar if tqdm is available
        if not isinstance(tbar, range):
            reg_value = current_logs["eval"].get("regularizer", "NA")
            postfix_str = (
                f"fitting_loss: {current_logs['eval']['fitting_loss']:.4f}, "
                f"regularizer: {reg_value:.4f},"
                f"total: {current_logs['eval']['total_loss']:.4f}"
            )
            tbar.set_postfix_str(postfix_str)

    def train(self, datamodule: ConditionalDataModule):
        logs = collections.defaultdict(lambda: collections.defaultdict(list))
        try:
            from tqdm import trange

            tbar = trange(self.num_train_iters, leave=True)
        except ImportError:
            tbar = range(self.num_train_iters)

        train_conditions = []
        grads = tree_map(jnp.zeros_like, self.state_neural_net.params)
        for step in tbar:
            is_logging_step = step % 100 == 0
            is_gradient_acc_step = (step + 1) % self.grad_acc_steps == 0
            train_batch, condition = self.generate_batch(datamodule, "train")
            valid_batch, _ = (
                ({"num_contexts": None}, None)
                if not is_logging_step
                else self.generate_batch(datamodule, "valid")
            )

            self.state_neural_net, grads, current_logs = self.step_fn(
                self.state_neural_net,
                grads=grads,
                train_batch=train_batch,
                valid_batch=valid_batch,
                is_logging_step=is_logging_step,
                is_gradient_acc_step=is_gradient_acc_step,
                n_train_contexts=train_batch["num_contexts"],
                n_valid_contexts=valid_batch["num_contexts"],
            )
            train_conditions.append(condition)

            if is_logging_step:
                self.update_logs(current_logs, logs, tbar)
        self.metrics["ott-logs"] = logs
        self.metrics["train_conditions"] = train_conditions

        return self.state_neural_net, logs

    def _get_step_fn(self) -> Callable:
        """Create a one step training and evaluation function."""

        def loss_fn(
            params: frozen_dict.FrozenDict,
            apply_fn: Callable,
            batch: Dict[str, jnp.ndarray],
            n_contexts: int,
        ) -> Tuple[float, Dict[str, float]]:
            """Loss function."""
            # map samples with the fitted map
            mapped_samples = apply_fn(
                {"params": params},
                batch["source"],
                batch["condition"],
                n_contexts,
            )

            # compute the loss
            val_fitting_loss = self.fitting_loss(batch["target"], mapped_samples)
            val_regularizer = self.regularizer(batch["source"], mapped_samples)
            val_tot_loss = val_fitting_loss + val_regularizer

            # store training logs
            loss_logs = {
                "total_loss": val_tot_loss,
                "fitting_loss": val_fitting_loss,
                "regularizer": val_regularizer,
            }

            return val_tot_loss, loss_logs

        @functools.partial(jax.jit, static_argnums=[4, 5, 6, 7])
        def step_fn(
            state_neural_net: train_state.TrainState,
            grads: frozen_dict.FrozenDict,
            train_batch: Dict[str, jnp.ndarray],
            valid_batch: Optional[Dict[str, jnp.ndarray]] = None,
            is_logging_step: bool = False,
            is_gradient_acc_step: bool = False,
            n_train_contexts: int = 2,
            n_valid_contexts: int = 2,
        ) -> Tuple[train_state.TrainState, frozen_dict.FrozenDict, Dict[str, float]]:
            """Step function."""
            # compute loss and gradients
            grad_fn = jax.value_and_grad(loss_fn, argnums=0, has_aux=True)
            (_, current_train_logs), step_grads = grad_fn(
                state_neural_net.params,
                state_neural_net.apply_fn,
                train_batch,
                n_train_contexts,
            )
            # Accumulate gradients
            grads = tree_map(lambda g, step_g: g + step_g, grads, step_grads)

            # logging step
            current_logs = {"train": current_train_logs, "eval": {}}
            if is_logging_step:
                _, current_eval_logs = loss_fn(
                    params=state_neural_net.params,
                    apply_fn=state_neural_net.apply_fn,
                    batch=valid_batch,
                    n_contexts=n_valid_contexts,
                )
                current_logs["eval"] = current_eval_logs

            # update state
            if is_gradient_acc_step:
                state_neural_net = state_neural_net.apply_gradients(
                    grads=tree_map(lambda g: g / self.grad_acc_steps, grads)
                )
                # Reset gradients
                grads = tree_map(jnp.zeros_like, grads)

            return state_neural_net, grads, current_logs

        return step_fn

    def transport(self, x, c, num_contexts):
        return self.state_neural_net.apply_fn(
            {"params": self.state_neural_net.params}, x, c, num_contexts
        )

    def evaluate(
        self,
        datamodule: ConditionalDataModule,
        identity: bool = False,
        n_samples: int = 9,
    ) -> None:
        """Evaluate a trained model on a validation set and save the metrics to a json file."""

        def evaluate_condition(
            loader_source: Iterator[jnp.ndarray],
            loader_target: Iterator[jnp.ndarray],
            cond_embeddings: jnp.ndarray,
            metrics: str,
            n_contexts,
        ):
            for enum, (source, target) in enumerate(zip(loader_source, loader_target)):
                if not identity:
                    transport = self.transport(source, cond_embeddings, n_contexts)
                else:
                    transport = source

                transport = datamodule.decoder(transport)
                target = datamodule.decoder(target)

                if datamodule.marker_idx:
                    target = target[:, datamodule.marker_idx]
                    transport = transport[:, datamodule.marker_idx]

                log_metrics(metrics, target, transport)
                if enum > n_samples:
                    break

        def evaluate_split(
            cond_to_loaders: Dict[
                str, Tuple[Iterator[jnp.ndarray], Iterator[jnp.ndarray]]
            ],
            split_type: str,
        ):
            self.metrics[split_type] = {}
            for cond, loader in cond_to_loaders.items():
                logger.info(f"Evaluation started on {cond} {split_type}.")
                cond_embedding, n_contexts = self.embedding_module(cond)
                loader_source, loader_target = loader

                self.metrics[split_type][cond] = {}
                init_logger_dict(
                    self.metrics[split_type][cond], datamodule.drug_condition
                )
                evaluate_condition(
                    loader_source,
                    loader_target,
                    cond_embedding,
                    self.metrics[split_type][cond],
                    n_contexts,
                )
                log_mean_metrics(self.metrics[split_type][cond])

        # Log in test set if present, otherwise valid otherwise train
        if self.datamodule.data_config.split[2] > 0:
            logger.info("Evaluating on test set")
            cond_to_loaders = datamodule.test_dataloaders()
            evaluate_split(
                cond_to_loaders=cond_to_loaders,
                split_type="test-set",
            )
        elif self.datamodule.data_config.split[1] > 0:
            logger.info("Evaluating on validation set")
            cond_to_loaders = datamodule.valid_dataloaders()
            evaluate_split(
                cond_to_loaders=cond_to_loaders,
                split_type="valid-set",
            )
        else:
            logger.info("Evaluating on train set")
            cond_to_loaders = datamodule.train_dataloaders()
            evaluate_split(
                cond_to_loaders=cond_to_loaders,
                split_type="train-set",
            )

        create_or_update_logfile(self.logger_path, self.metrics)

    @property
    def model(self) -> nn.Module:
        return self.state_neural_net

    @model.setter
    def model(self, value: nn.Module):
        """Setter for the model to be checkpointed."""
        self.state_neural_net = value



import jax
import jax.numpy as jnp
import numpy as np
from ott.geometry import costs
from ott.geometry.pointcloud import PointCloud
from ott.neural.methods.monge_gap import monge_gap_from_samples
from ott.solvers.linear import solve
from ott.tools.sinkhorn_divergence import sinkhorn_divergence
from sklearn.metrics.pairwise import rbf_kernel


def average_r2(target: jnp.ndarray, transport: jnp.ndarray) -> float:
    """
    Calculate the correlation coefficient r^2 between the means of average features in target and tansport.
    """
    target_means = jnp.mean(target, axis=0)
    transport_means = jnp.mean(transport, axis=0)
    average_r2 = np.corrcoef(target_means, transport_means)[0, 1] ** 2
    return float(average_r2)


def drug_signature(target: jnp.ndarray, transport: jnp.ndarray) -> float:
    """Calculates the euclidien distance between the marginal means of the target and transported measures."""
    target_means = jnp.mean(target, 0)
    transport_means = jnp.mean(transport, 0)
    return float(jnp.linalg.norm(target_means - transport_means))


def maximum_mean_discrepancy(
    target: jnp.ndarray, transport: jnp.ndarray, gamma: float
) -> float:
    """Calculates the maximum mean discrepancy between two measures."""
    xx = rbf_kernel(target, target, gamma)
    xy = rbf_kernel(target, transport, gamma)
    yy = rbf_kernel(transport, transport, gamma)

    return float(xx.mean() + yy.mean() - 2 * xy.mean())


def compute_scalar_mmd(
    target: jnp.ndarray,
    transport: jnp.ndarray,
    gammas: list[float] = [2, 1, 0.5, 0.1, 0.01, 0.005],
):
    """
    Calculates the maximum mean discrepancy between the target
    and the transported measures, using gaussian kernel,averaging for different gammas.
    """

    def safe_mmd(*args):
        try:
            mmd = maximum_mean_discrepancy(*args)
        except ValueError:
            mmd = jnp.nan
        return mmd

    return float(np.mean(list(map(lambda x: safe_mmd(target, transport, x), gammas))))


def wasserstein_distance(
    target: jnp.ndarray, transport: jnp.ndarray, epsilon: float = 0.1
) -> float:
    """
    Calculates the Wasserstain distance between two measures
    using the Sinkhorn algorithm on the regularized OT formulation.
    """
    geom = PointCloud(target, transport, cost_fn=costs.Euclidean(), epsilon=epsilon)
    solver = jax.jit(solve)
    ot = solver(geom)
    return ot.reg_ot_cost


def fitting_loss(
    target: jnp.ndarray, transport: jnp.ndarray, epsilon_fitting: float
) -> float:
    """Calculates the sinkhorn divergence between two measures."""
    divergence, output_comp = sinkhorn_divergence(
        PointCloud,
        target,
        transport,
        cost_fn=costs.Euclidean(),
        epsilon=epsilon_fitting,
    )
    return divergence


def sinkhorn_div(target: jnp.ndarray, transport: jnp.ndarray) -> float:
    """Calculates the sinkhorn divergence between two measures."""
    return fitting_loss(target, transport, 0.1)


def regularizer(
    target: jnp.ndarray,
    transport: jnp.ndarray,
    epsilon_regularizer: float,
    cost: str,
):
    """Calculates the Monge Gap between two measures."""
    cost_fn = cost_factory[cost]
    gap = monge_gap_from_samples(
        target,
        transport,
        cost_fn=cost_fn,
        epsilon=epsilon_regularizer,
        return_output=False,
    )
    return gap


def eucledian_monge_gap(target: jnp.ndarray, transport: jnp.ndarray) -> float:
    return regularizer(target, transport, 1, "euclidean")


cost_factory = {"euclidean": costs.Euclidean()}



from pathlib import Path

import pytest

from cmonge.datasets.single_loader import SciPlexModule
from cmonge.utils import load_config


@pytest.fixture(scope="session")
def synthetic_config():
    config_path = Path("cmonge/tests/configs/synthetic.yml")
    config = load_config(config_path)
    return config


@pytest.fixture(scope="module")
def synthetic_data(synthetic_config):
    module = SciPlexModule(synthetic_config.data)
    return module
